# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import os
import time
from typing import Dict, List, NamedTuple, Set, Tuple

import psutil
import xoscar as mo
from xoscar.backends.allocate_strategy import IdleLabel
from xoscar.errors import NoFreeSlot, SlotStateError

from ....resource import Resource
from ....typing import BandType
from ...cluster import ClusterAPI, WorkerSlotInfo

logger = logging.getLogger(__name__)


class DispatchDumpType(NamedTuple):
    free_slots: Set
    fresh_slots: Set


class WorkerSlotManagerActor(mo.Actor):
    _band_slot_infos: Dict[str, List[WorkerSlotInfo]]

    def __init__(self):
        self._cluster_api = None
        self._global_resource_ref = None

        self._band_slot_managers = dict()  # type: Dict[str, mo.ActorRef]

    async def __post_create__(self):
        self._cluster_api = await ClusterAPI.create(self.address)

        band_to_resource = await self._cluster_api.get_bands()
        for band, resource in band_to_resource.items():
            self._band_slot_managers[band] = await mo.create_actor(
                BandSlotManagerActor,
                band,
                int(resource.num_cpus or resource.num_gpus),
                self._global_resource_ref,
                uid=BandSlotManagerActor.gen_uid(band[1]),
                address=self.address,
            )

    async def __pre_destroy__(self):
        await asyncio.gather(
            *[mo.destroy_actor(ref) for ref in self._band_slot_managers.values()]
        )


class BandSlotManagerActor(mo.Actor):
    _free_slots: Set[int]
    _fresh_slots: Set[int]

    @classmethod
    def gen_uid(cls, band_name: str):
        return f"{band_name}_band_slot_manager"

    def __init__(
        self, band: BandType, n_slots: int, global_resource_ref: mo.ActorRef = None
    ):
        super().__init__()
        self._cluster_api = None

        self._band = band
        self._band_name = band[1]
        self._global_resource_ref = global_resource_ref
        self._n_slots = n_slots

        self._semaphore = asyncio.Semaphore(0)
        self._slot_control_refs = dict()
        self._free_slots = set()
        self._fresh_slots = set()
        self._slot_kill_events = dict()

        self._restarting = False
        self._restart_done_event = asyncio.Event()

        self._session_stid_to_slot = dict()
        self._slot_to_session_stid = dict()
        self._last_report_time = time.time()

        self._slot_to_proc = dict()
        self._usage_upload_task = None

    async def __post_create__(self):
        try:
            self._cluster_api = await ClusterAPI.create(self.address)
        except mo.ActorNotExist:
            pass

        strategy = IdleLabel(self._band_name, f"worker_slot_control")
        for slot_id in range(self._n_slots):
            self._slot_control_refs[slot_id] = await mo.create_actor(
                BandSlotControlActor,
                self.ref(),
                self._band_name,
                slot_id,
                uid=BandSlotControlActor.gen_uid(self._band_name, slot_id),
                address=self.address,
                allocate_strategy=strategy,
            )
            self._fresh_slots.add(slot_id)

        self._upload_slot_usage_with_delay()

    async def __pre_destroy__(self):
        self._usage_upload_task.cancel()

    async def _get_global_resource_ref(self):
        if self._global_resource_ref is not None:
            return self._global_resource_ref

        from ..supervisor import GlobalResourceManagerActor

        try:
            [self._global_resource_ref] = await self._cluster_api.get_supervisor_refs(
                [GlobalResourceManagerActor.default_uid()]
            )
        except mo.ActorNotExist:
            self._global_resource_ref = None
        return self._global_resource_ref

    def get_slot_address(self, slot_id: int):
        return self._slot_control_refs[slot_id].address

    def get_subtask_slot(self, session_stid: Tuple[str, str]):
        return self._session_stid_to_slot.get(session_stid)

    async def acquire_free_slot(self, session_stid: Tuple[str, str], block=True):
        if not block and self._semaphore.locked():
            raise NoFreeSlot(f"No free slot for {session_stid}")
        yield self._semaphore.acquire()
        if self._restarting:
            yield self._restart_done_event.wait()

        slot_id = self._free_slots.pop()
        self._fresh_slots.difference_update([slot_id])
        self._slot_to_session_stid[slot_id] = session_stid
        self._session_stid_to_slot[session_stid] = slot_id
        logger.debug("Slot %d acquired for subtask %r", slot_id, session_stid)
        raise mo.Return(slot_id)

    def release_free_slot(self, slot_id: int, session_stid: Tuple[str, str]):
        acquired_session_stid = self._slot_to_session_stid.pop(slot_id, None)
        if acquired_session_stid is None:
            raise SlotStateError(f"Slot {slot_id} is not acquired.")
        if acquired_session_stid != session_stid:
            raise SlotStateError(
                f"Slot {slot_id} releasing state incorrect, "
                f"the acquired session_stid: {acquired_session_stid}, "
                f"the releasing session_stid: {session_stid}"
            )
        acquired_slot_id = self._session_stid_to_slot.pop(acquired_session_stid)
        assert (
            acquired_slot_id == slot_id
        ), f"{acquired_session_stid}: acquired_slot_id {acquired_slot_id} != slot_id {slot_id}"

        logger.debug("Slot %d released", slot_id)

        if slot_id not in self._free_slots:
            self._free_slots.add(slot_id)
            self._semaphore.release()

    def register_slot(self, slot_id: int, pid: int):
        try:
            self._fresh_slots.add(slot_id)
            if slot_id in self._slot_kill_events:
                event = self._slot_kill_events.pop(slot_id)
                event.set()
            if slot_id in self._slot_to_session_stid:
                # We should release the slot by one role, if the slot is
                # acquired by the SubtaskExecutionActor, then the slot
                # should be released by it, too.
                session_stid = self._slot_to_session_stid[slot_id]
                logger.info(
                    "Slot %s registered by pid %s, current acquired session_stid is %s",
                    slot_id,
                    pid,
                    session_stid,
                )
            else:
                if slot_id not in self._free_slots:
                    self._free_slots.add(slot_id)
                    self._semaphore.release()
        finally:
            # psutil may raises exceptions, but currently we can't handle the register exception,
            # so put it to the finally.
            # TODO(fyrestone): handle register_slot failure.
            self._slot_to_proc[slot_id] = proc = psutil.Process(pid)
            # collect initial stats for the process
            proc.cpu_percent(interval=None)

    async def _kill_slot(self, slot_id: int):
        if slot_id in self._slot_kill_events:
            await self._slot_kill_events[slot_id].wait()
            return

        event = self._slot_kill_events[slot_id] = asyncio.Event()
        # TODO(fyrestone): Make it more reliable. e.g. kill_actor
        # success but the actor does not restart.
        try:
            await mo.kill_actor(self._slot_control_refs[slot_id])
        except ConnectionError:
            pass
        await event.wait()

    async def kill_slot(self, slot_id: int):
        self._free_slots.difference_update([slot_id])
        yield self._kill_slot(slot_id)

    async def restart_free_slots(self):
        if self._restarting:
            yield self._restart_done_event.wait()
            return

        self._restart_done_event = asyncio.Event()
        self._restarting = True
        slot_ids = [
            slot_id for slot_id in self._free_slots if slot_id not in self._fresh_slots
        ]
        if slot_ids:
            yield asyncio.gather(*[self._kill_slot(slot_id) for slot_id in slot_ids])
            logger.info("%d idle slots restarted", len(slot_ids))

        self._restarting = False
        self._restart_done_event.set()

    def _upload_slot_usage_with_delay(self, delay: int = 1):
        self._usage_upload_task = self.ref().upload_slot_usages.tell_delay(
            periodical=True, delay=delay
        )

    async def upload_slot_usages(self, periodical: bool = False):
        delays = []
        slot_infos = []
        global_resource_ref = await self._get_global_resource_ref()

        if global_resource_ref is None:  # pragma: no cover
            if periodical:
                self._upload_slot_usage_with_delay()
            return

        for slot_id, proc in self._slot_to_proc.items():
            if slot_id not in self._slot_to_session_stid:
                continue
            session_id, subtask_id = self._slot_to_session_stid[slot_id]
            cpu_usage, gpu_usage, processor_usage = 0, 0, 0
            if self._band_name.startswith("gpu"):
                processor_usage = gpu_usage = 1
            else:
                try:
                    processor_usage = cpu_usage = (
                        proc.cpu_percent(interval=None) / 100.0
                    )
                except psutil.NoSuchProcess:  # pragma: no cover
                    continue
                except psutil.AccessDenied as e:  # pragma: no cover
                    logger.warning("Access denied when getting cpu percent: %s", e)
                    processor_usage = cpu_usage = 0.0

            slot_infos.append(
                WorkerSlotInfo(
                    slot_id=slot_id,
                    session_id=session_id,
                    subtask_id=subtask_id,
                    processor_usage=processor_usage,
                )
            )

            if global_resource_ref is not None:  # pragma: no branch
                # FIXME fix band slot mistake
                delays.append(
                    global_resource_ref.update_subtask_resources.delay(
                        self._band[1],
                        session_id,
                        subtask_id,
                        Resource(
                            num_cpus=max(1.0, cpu_usage), num_gpus=max(1.0, gpu_usage)
                        ),
                    )
                )

        if delays:  # pragma: no branch
            yield global_resource_ref.update_subtask_resources.batch(*delays)
        if self._cluster_api is not None:
            await self._cluster_api.set_band_slot_infos(self._band_name, slot_infos)

        if periodical:
            self._upload_slot_usage_with_delay()

    def dump_data(self):
        """
        Get all refs of slots of a queue
        """
        return DispatchDumpType(self._free_slots, self._fresh_slots)


class BandSlotControlActor(mo.Actor):
    @classmethod
    def gen_uid(cls, band_name: str, slot_id: int):
        return f"{band_name}_{slot_id}_band_slot_control"

    def __init__(self, manager_ref, band_name, slot_id: int):
        self._manager_ref = manager_ref
        self._band_name = band_name
        self._slot_id = slot_id
        self._report_task = None

    async def __post_create__(self):
        self._report_task = asyncio.create_task(self._report_slot_ready())

    async def _report_slot_ready(self):
        from ...cluster.api import ClusterAPI

        try:
            self._cluster_api = await ClusterAPI.create(self.address)
            await self._cluster_api.wait_node_ready()
        except mo.ActorNotExist:
            pass

        await mo.wait_actor_pool_recovered(self.address)
        await self._manager_ref.register_slot.tell(self._slot_id, os.getpid())



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import itertools
from collections import namedtuple

import numpy as np
import pandas as pd

from ....dataframe import DataFrame as MarsDataFrame
from ....dataframe import Series as MarsSeries
from ....tensor import tensor as mars_tensor


class LGBMModelType(enum.Enum):
    CLASSIFIER = 0
    REGRESSOR = 1
    RANKER = 2


_model_type_to_model = dict()


def get_model_cls_from_type(model_type: LGBMModelType):
    import lightgbm

    if not _model_type_to_model:
        _model_type_to_model.update(
            {
                LGBMModelType.CLASSIFIER: lightgbm.LGBMClassifier,
                LGBMModelType.REGRESSOR: lightgbm.LGBMRegressor,
                LGBMModelType.RANKER: lightgbm.LGBMRanker,
            }
        )
    return _model_type_to_model[model_type]


TrainTuple = namedtuple("TrainTuple", "data label sample_weight init_score")


class LGBMScikitLearnBase:
    def __init__(self, *args, **kwargs):
        if args and isinstance(args[0], self._get_lgbm_class()):
            model = args[0]
            super().__init__(**model.get_params())
            self._copy_extra_params(model, self)
        else:
            super().__init__(*args, **kwargs)

    @classmethod
    def _get_lgbm_class(cls):
        try:
            return getattr(cls, "_lgbm_class")
        except AttributeError:
            lgbm_class = next(
                base for base in cls.__bases__ if base.__module__.startswith("lightgbm")
            )
            cls._lgbm_class = lgbm_class
            return lgbm_class

    @classmethod
    def _get_param_names(cls):
        return cls._get_lgbm_class()._get_param_names()

    @staticmethod
    def _copy_extra_params(source, dest):
        params = source.get_params()
        attributes = source.__dict__
        extra_param_names = set(attributes.keys()).difference(params.keys())
        for name in extra_param_names:
            setattr(dest, name, attributes[name])

    @staticmethod
    def _convert_tileable(obj):
        if isinstance(obj, np.ndarray):
            return mars_tensor(obj)
        elif isinstance(obj, pd.DataFrame):
            return MarsDataFrame(obj)
        elif isinstance(obj, pd.Series):
            return MarsSeries(obj)
        return obj

    @classmethod
    def _wrap_train_tuple(cls, data, label, sample_weight=None, init_score=None):
        data = cls._convert_tileable(data)
        label = cls._convert_tileable(label)
        sample_weight = cls._convert_tileable(sample_weight)
        init_score = cls._convert_tileable(init_score)
        return TrainTuple(data, label, sample_weight, init_score)

    @staticmethod
    def _wrap_eval_tuples(eval_set=None, eval_sample_weight=None, eval_init_score=None):
        if not eval_set:
            return None

        tps = []
        for (data, label), weight, score in zip(
            eval_set,
            eval_sample_weight or itertools.repeat(None),
            eval_init_score or itertools.repeat(None),
        ):
            tps.append(TrainTuple(data, label, weight, score))
        return tps

    def fit(self, X, y, sample_weight=None, **kwargs):
        raise NotImplementedError

    def predict(self, X, **kwargs):
        raise NotImplementedError

    def predict_proba(self, X, **kwargs):
        raise NotImplementedError

    def load_model(self, model):
        self.set_params(**self.get_params())
        self._copy_extra_params(model, self)
        return self



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
from typing import Any, Union

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import OutputType
from ...core.context import Context
from ...core.operand import MapReduceOperand, OperandStage
from ...serialization.serializables import AnyField, Int32Field, ListField
from ..operands import DataFrameOperandMixin, DataFrameShuffleProxy
from ..utils import build_concatenated_rows_frame, hash_dataframe_on, parse_index


class DataFramePivot(MapReduceOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.PIVOT

    columns = AnyField("columns", default=None)
    index = AnyField("index", default=None)
    values = AnyField("values", default=None)
    shuffle_size = Int32Field("shuffle_size")
    output_columns = ListField("output_columns", default=None)

    @classmethod
    def execute_map(cls, ctx: Union[dict, Context], op: "DataFramePivot"):
        chunk = op.outputs[0]
        df = ctx[op.inputs[0].key]

        filters = hash_dataframe_on(df, op.index, op.shuffle_size)

        for index_idx, index_filter in enumerate(filters):
            reducer_index = (index_idx, chunk.index[1])
            ctx[chunk.key, reducer_index] = (
                ctx.get_current_chunk().index,
                df.iloc[index_filter],
            )

    @classmethod
    def execute_reduce(cls, ctx: Union[dict, Context], op: "DataFramePivot"):
        chunk = op.outputs[0]
        input_idx_to_df = dict(op.iter_mapper_data(ctx))
        row_idxes = sorted(input_idx_to_df.keys())

        res = []
        for row_idx in row_idxes:
            row_df = input_idx_to_df.get(row_idx, None)
            if row_df is not None:
                res.append(row_df)
        ctx[chunk.key] = pd.concat(res, axis=0)

    @classmethod
    def execute_combine(cls, ctx: Union[dict, Context], op: "DataFramePivot"):
        input_data = ctx[op.inputs[0].key]
        out = op.outputs[0]

        for dtype in op.output_columns:
            if dtype not in input_data.dtypes.index:
                input_data[dtype] = np.nan

        ctx[out.key] = input_data

    @classmethod
    def execute(cls, ctx: Union[dict, Context], op: "DataFramePivot"):
        if op.stage == OperandStage.map:
            cls.execute_map(ctx, op)
        elif op.stage == OperandStage.reduce:
            cls.execute_reduce(ctx, op)
        elif op.stage == OperandStage.combine:
            cls.execute_combine(ctx, op)
        else:
            input_data = ctx[op.inputs[0].key]
            out = op.outputs[0]
            kwargs = {}
            if op.index is not None:
                kwargs["index"] = op.index
            if op.values is not None:
                kwargs["values"] = op.values
            result = input_data.pivot(columns=op.columns, **kwargs)
            ctx[out.key] = result

    @classmethod
    def tile_one_chunk(cls, op: "DataFramePivot"):
        in_df = op.inputs[0]
        out_df = op.outputs[0]

        chunks = []
        for c in in_df.chunks:
            new_op = op.copy().reset_key()
            new_op.tileable_op_key = op.key
            chunks.append(
                new_op.new_chunk(
                    [c],
                    shape=(np.nan, np.nan),
                    index=c.index,
                    dtypes=out_df.dtypes,
                    index_value=c.index_value,
                    columns_value=c.columns_value,
                )
            )

        new_op = op.copy()
        kw = out_df.params.copy()
        new_nsplits = ((np.nan,), (np.nan,))
        kw.update(dict(chunks=chunks, nsplits=new_nsplits))
        return new_op.new_tileables(op.inputs, **kw)

    @classmethod
    def tile(cls, op: "DataFramePivot"):
        in_df = build_concatenated_rows_frame(op.inputs[0])

        if len(in_df.chunks) == 1:
            return cls.tile_one_chunk(op)

        out_df = op.outputs[0]
        output_type = OutputType.dataframe
        chunk_shape = in_df.chunk_shape

        # generate map chunks
        map_chunks = []
        for chunk in in_df.chunks:
            map_op = op.copy().reset_key()
            map_op.stage = OperandStage.map
            map_op.shuffle_size = chunk_shape[0]
            map_op._output_types = [output_type]
            chunk_inputs = [chunk]
            map_chunks.append(
                map_op.new_chunk(
                    chunk_inputs,
                    shape=(np.nan, np.nan),
                    index=chunk.index,
                )
            )
        proxy_chunk = DataFrameShuffleProxy(output_types=[output_type]).new_chunk(
            map_chunks, shape=()
        )
        # generate reduce chunks
        reduce_chunks = []
        out_indices = list(itertools.product(*(range(s) for s in chunk_shape)))
        for ordinal, out_idx in enumerate(out_indices):
            reduce_op = op.copy().reset_key()
            reduce_op._output_types = [output_type]
            reduce_op.stage = OperandStage.reduce
            reduce_op.reducer_ordinal = ordinal
            reduce_op.n_reducers = len(out_indices)
            reduce_chunks.append(
                reduce_op.new_chunk(
                    [proxy_chunk], shape=(np.nan, np.nan), index=out_idx
                )
            )

        out_chunks = []
        for chunk in reduce_chunks:
            new_op = op.copy().reset_key()
            new_shape = (np.nan, np.nan)

            params = dict(shape=new_shape, index=chunk.index)
            params.update(
                dict(
                    dtypes=in_df.dtypes,
                    columns_value=in_df.columns_value,
                    index_value=parse_index(None, chunk.key, proxy_chunk.key),
                )
            )
            out_chunks.append(new_op.new_chunk([chunk], **params))

        yield out_chunks

        filtered_chunks = [chunk for chunk in out_chunks if chunk.shape[1] != 0]
        for i, chunk in enumerate(filtered_chunks):
            chunk._index = (i, 0)

        combined_dtypes = pd.concat(
            [chunk.dtypes for chunk in filtered_chunks]
        ).to_dict()
        output_columns = list(combined_dtypes.keys())
        output_dtypes = pd.Series(combined_dtypes)

        # generate combine chunks
        combine_chunks = []
        for chunk in filtered_chunks:
            if chunk.shape[1] < len(output_columns):
                combine_op = op.copy().reset_key()
                combine_op.stage = OperandStage.combine
                combine_op.output_columns = output_columns
                params = dict(
                    shape=(chunk.shape[0], len(output_columns)), index=chunk.index
                )
                params.update(
                    dict(
                        dtypes=output_dtypes,
                        columns_value=in_df.columns_value,
                        index_value=chunk.index_value,
                    )
                )
                combine_chunks.append(combine_op.new_chunk([chunk], **params))
            else:
                combine_chunks.append(chunk)

        new_op = op.copy()
        kw = out_df.params.copy()
        new_nsplits = (
            tuple(chunk.shape[0] for chunk in combine_chunks),
            (len(output_columns),),
        )
        kw.update(dict(chunks=combine_chunks, nsplits=new_nsplits))

        return new_op.new_tileables(op.inputs, **kw)

    def __call__(self, df):
        return self.new_dataframe([df])


def df_pivot(
    df: pd.DataFrame, columns: Any, *, index: Any = None, values: Any = None
) -> pd.DataFrame:
    """
    Return reshaped DataFrame organized by given index / column values.

    Reshape data (produce a "pivot" table) based on column values. Uses
    unique values from specified `index` / `columns` to form axes of the
    resulting DataFrame. This function does not support data
    aggregation, multiple values will result in a MultiIndex in the
    columns. See the :ref:`User Guide <reshaping>` for more on reshaping.

    Parameters
    ----------
    columns : str or object or a list of str
        Column to use to make new frame's columns.

        .. versionchanged:: 1.1.0
            Also accept list of columns names.

    index : str or object or a list of str, optional
        Column to use to make new frame's index. If not given, uses existing index.

        .. versionchanged:: 1.1.0
            Also accept list of index names.

    values : str, object or a list of the previous, optional
        Column(s) to use for populating new frame's values. If not
        specified, all remaining columns will be used and the result will
        have hierarchically indexed columns.

    Returns
    -------
    DataFrame
        Returns reshaped DataFrame.

    Raises
    ------
    ValueError:
        When there are any `index`, `columns` combinations with multiple
        values. `DataFrame.pivot_table` when you need to aggregate.

    See Also
    --------
    DataFrame.pivot_table : Generalization of pivot that can handle
        duplicate values for one index/column pair.
    DataFrame.unstack : Pivot based on the index values instead of a
        column.
    wide_to_long : Wide panel to long format. Less flexible but more
        user-friendly than melt.

    Notes
    -----
    For finer-tuned control, see hierarchical indexing documentation along
    with the related stack/unstack methods.

    Reference :ref:`the user guide <reshaping.pivot>` for more examples.

    Examples
    --------
    >>> df = pd.DataFrame({'foo': ['one', 'one', 'one', 'two', 'two',
    ...                            'two'],
    ...                    'bar': ['A', 'B', 'C', 'A', 'B', 'C'],
    ...                    'baz': [1, 2, 3, 4, 5, 6],
    ...                    'zoo': ['x', 'y', 'z', 'q', 'w', 't']})
    >>> df
        foo   bar  baz  zoo
    0   one   A    1    x
    1   one   B    2    y
    2   one   C    3    z
    3   two   A    4    q
    4   two   B    5    w
    5   two   C    6    t

    >>> df.pivot(index='foo', columns='bar', values='baz')
    bar  A   B   C
    foo
    one  1   2   3
    two  4   5   6

    >>> df.pivot(index='foo', columns='bar')['baz']
    bar  A   B   C
    foo
    one  1   2   3
    two  4   5   6

    >>> df.pivot(index='foo', columns='bar', values=['baz', 'zoo'])
            baz       zoo
    bar   A  B  C   A  B  C
    foo
    one   1  2  3   x  y  z
    two   4  5  6   q  w  t

    You could also assign a list of column names or a list of index names.

    >>> df = pd.DataFrame({
    ...        "lev1": [1, 1, 1, 2, 2, 2],
    ...        "lev2": [1, 1, 2, 1, 1, 2],
    ...        "lev3": [1, 2, 1, 2, 1, 2],
    ...        "lev4": [1, 2, 3, 4, 5, 6],
    ...        "values": [0, 1, 2, 3, 4, 5]})
    >>> df
        lev1 lev2 lev3 lev4 values
    0   1    1    1    1    0
    1   1    1    2    2    1
    2   1    2    1    3    2
    3   2    1    2    4    3
    4   2    1    1    5    4
    5   2    2    2    6    5

    >>> df.pivot(index="lev1", columns=["lev2", "lev3"], values="values")
    lev2    1         2
    lev3    1    2    1    2
    lev1
    1     0.0  1.0  2.0  NaN
    2     4.0  3.0  NaN  5.0

    >>> df.pivot(index=["lev1", "lev2"], columns=["lev3"], values="values")
            lev3    1    2
    lev1  lev2
        1     1  0.0  1.0
                2  2.0  NaN
        2     1  4.0  3.0
                2  NaN  5.0

    A ValueError is raised if there are any duplicates.

    >>> df = pd.DataFrame({"foo": ['one', 'one', 'two', 'two'],
    ...                    "bar": ['A', 'A', 'B', 'C'],
    ...                    "baz": [1, 2, 3, 4]})
    >>> df
        foo bar  baz
    0  one   A    1
    1  one   A    2
    2  two   B    3
    3  two   C    4

    Notice that the first two rows are the same for our `index`
    and `columns` arguments.

    >>> df.pivot(index='foo', columns='bar', values='baz')
    Traceback (most recent call last):
        ...
    ValueError: Index contains duplicate entries, cannot reshape
    """
    op = DataFramePivot(columns=columns, index=index, values=values)
    return op(df)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import os
import uuid
from typing import Dict, List, Optional, Tuple

from xoscar.serialization import AioDeserializer, AioSerializer

from ..lib.aio import AioFilesystem
from ..lib.filesystem import FileSystem, LocalFileSystem, get_fs
from ..utils import implements, mod_hash
from .base import ObjectInfo, StorageBackend, StorageLevel, register_storage_backend
from .core import StorageFileObject


@register_storage_backend
class FileSystemStorage(StorageBackend):
    name = "filesystem"

    def __init__(
        self, fs: FileSystem, root_dirs: List[str], level: StorageLevel, size: int
    ):
        self._fs = AioFilesystem(fs)
        self._root_dirs = root_dirs
        self._level = level
        self._size = size

    @classmethod
    @implements(StorageBackend.setup)
    async def setup(cls, **kwargs) -> Tuple[Dict, Dict]:
        root_dirs = kwargs.pop("root_dirs")
        level = kwargs.pop("level")
        size = kwargs.pop("size", None)
        fs = kwargs.pop("fs", None)
        if kwargs:  # pragma: no cover
            raise TypeError(
                f'FileSystemStorage got unexpected config: {",".join(kwargs)}'
            )

        if isinstance(root_dirs, str):
            root_dirs = root_dirs.split(":")
        if isinstance(level, str):
            level = StorageLevel.from_str(level)

        if fs is None:
            fs = get_fs(root_dirs[0])

        for d in root_dirs:
            if not fs.exists(d):
                fs.mkdir(d)
        params = dict(fs=fs, root_dirs=root_dirs, level=level, size=size)
        return params, params

    @staticmethod
    @implements(StorageBackend.teardown)
    async def teardown(**kwargs):
        fs = kwargs.get("fs")
        root_dirs = kwargs.get("root_dirs")
        for d in root_dirs:
            fs.delete(d, recursive=True)

    @property
    @implements(StorageBackend.level)
    def level(self) -> StorageLevel:
        return self._level

    @property
    @implements(StorageBackend.size)
    def size(self) -> Optional[int]:
        return self._size

    def _generate_path(self):
        file_name = str(uuid.uuid4())
        selected_index = mod_hash(file_name, len(self._root_dirs))
        selected_dir = self._root_dirs[selected_index]
        return os.path.join(selected_dir, file_name)

    @implements(StorageBackend.get)
    async def get(self, object_id, **kwargs) -> object:
        if kwargs:  # pragma: no cover
            raise NotImplementedError(f'Got unsupported args: {",".join(kwargs)}')

        file = await self._fs.open(object_id, "rb")
        async with file as f:
            deserializer = AioDeserializer(f)
            return await deserializer.run()

    @implements(StorageBackend.put)
    async def put(self, obj, importance: int = 0) -> ObjectInfo:
        serializer = AioSerializer(obj)
        buffers = await serializer.run()
        buffer_size = sum(getattr(buf, "nbytes", len(buf)) for buf in buffers)

        path = self._generate_path()
        file = await self._fs.open(path, "wb")
        async with file as f:
            for buffer in buffers:
                await f.write(buffer)

        return ObjectInfo(size=buffer_size, object_id=path)

    @implements(StorageBackend.delete)
    async def delete(self, object_id):
        await self._fs.delete(object_id)

    @implements(StorageBackend.list)
    async def list(self) -> List:
        file_list = []
        for d in self._root_dirs:
            file_list.extend(list(await self._fs.ls(d)))
        return file_list

    @implements(StorageBackend.object_info)
    async def object_info(self, object_id) -> ObjectInfo:
        stat = await self._fs.stat(object_id)
        return ObjectInfo(size=stat["size"], object_id=object_id)

    @implements(StorageBackend.open_writer)
    async def open_writer(self, size=None) -> StorageFileObject:
        path = self._generate_path()
        file = await self._fs.open(path, "wb")
        return StorageFileObject(file, file.name)

    @implements(StorageBackend.open_reader)
    async def open_reader(self, object_id) -> StorageFileObject:
        file = await self._fs.open(object_id, "rb")
        return StorageFileObject(file, file.name)


@register_storage_backend
class DiskStorage(FileSystemStorage):
    name = "disk"

    @classmethod
    @implements(StorageBackend.setup)
    async def setup(cls, **kwargs) -> Tuple[Dict, Dict]:
        kwargs["level"] = StorageLevel.DISK
        return await super().setup(**kwargs)


@register_storage_backend
class AlluxioStorage(FileSystemStorage):
    name = "alluxio"

    def __init__(
        self,
        root_dir: str,
        local_environ: bool,  # local_environ means standalone mode
        level: StorageLevel = None,
        size: int = None,
    ):
        self._fs = AioFilesystem(LocalFileSystem())
        self._root_dirs = [root_dir]
        self._level = level
        self._size = size
        self._local_environ = local_environ

    @classmethod
    @implements(StorageBackend.setup)
    async def setup(cls, **kwargs) -> Tuple[Dict, Dict]:
        kwargs["level"] = StorageLevel.MEMORY
        root_dir = kwargs.get("root_dir")
        local_environ = kwargs.get("local_environ")
        if local_environ:
            proc = await asyncio.create_subprocess_shell(
                f"""$ALLUXIO_HOME/bin/alluxio fs mkdir /alluxio-storage
                $ALLUXIO_HOME/integration/fuse/bin/alluxio-fuse mount {root_dir} /alluxio-storage
                """
            )
            await proc.wait()
        params = dict(
            root_dir=root_dir,
            level=StorageLevel.MEMORY,
            size=None,
            local_environ=local_environ,
        )
        return params, dict(root_dir=root_dir)

    @staticmethod
    @implements(StorageBackend.teardown)
    async def teardown(**kwargs):
        root_dir = kwargs.get("root_dir")
        proc = await asyncio.create_subprocess_shell(
            f"""$ALLUXIO_HOME/integration/fuse/bin/alluxio-fuse unmount {root_dir} /alluxio-storage
            $ALLUXIO_HOME/bin/alluxio fs rm -R /alluxio-storage
            """
        )
        await proc.wait()


@register_storage_backend
class JuiceFSStorage(FileSystemStorage):
    name = "juicefs"

    def __init__(
        self,
        root_dirs: List[str],
        in_k8s: bool = False,
        local_environ: bool = False,  # local_environ means standalone mode
        level: StorageLevel = None,
        size: int = None,
    ):
        self._root_dirs = root_dirs
        self._in_k8s = in_k8s
        if not self._in_k8s:
            self._fs = AioFilesystem(LocalFileSystem())
            self._level = level
            self._size = size
            self._local_environ = local_environ

    @classmethod
    @implements(StorageBackend.setup)
    async def setup(cls, **kwargs) -> Tuple[Dict, Dict]:
        kwargs["level"] = StorageLevel.MEMORY
        in_k8s = kwargs.get("in_k8s")
        root_dirs = kwargs.get("root_dirs")
        params = dict(
            root_dirs=root_dirs,
            level=StorageLevel.MEMORY,
            size=None,
        )
        if not in_k8s:
            local_environ = kwargs.get("local_environ")
            metadata_url = kwargs.get("metadata_url", None)
            if metadata_url is None:
                raise ValueError(
                    "For external storage JuiceFS, you must specify the metadata url for its metadata storage, for example 'redis://172.17.0.5:6379/1'."
                )
            if local_environ:
                proc = await asyncio.create_subprocess_shell(
                    f"""juicefs format {metadata_url} jfs
                    juicefs mount {metadata_url} {root_dirs[0]} -d
                    """
                )
                await proc.wait()
            params["local_environ"] = local_environ
        return params, dict(root_dirs=root_dirs)

    @staticmethod
    @implements(StorageBackend.teardown)
    async def teardown(**kwargs):
        in_k8s = kwargs.get("in_k8s")
        if not in_k8s:
            root_dirs = kwargs.get("root_dirs")
            proc = await asyncio.create_subprocess_shell(
                f"""juicefs umount {root_dirs[0]}
                """
            )
            await proc.wait()



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Union

import numpy as np

from ... import opcodes
from ...core import OutputType, get_output_types, recursive_tile
from ...serialization.serializables import (
    AnyField,
    Int16Field,
    Int64Field,
    KeyField,
    StringField,
)
from ...typing import TileableType
from ..align import (
    align_dataframe_dataframe,
    align_dataframe_series,
    align_series_series,
)
from ..core import IndexValue
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_empty_df, parse_index, validate_axis


class _NoNeedToAlign(Exception):
    pass


class DataFrameAlign(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = opcodes.ALIGN

    lhs = KeyField("lhs")
    rhs = KeyField("rhs")
    join = StringField("join", default=None)
    axis = Int16Field("axis", default=None)
    level = AnyField("level", default=None)
    fill_value = AnyField("fill_value", default=None)
    method = StringField("method", default=None)
    limit = Int64Field("limit", default=None)
    fill_axis = Int16Field("fill_axis", default=None)
    broadcast_axis = Int16Field("broadcast_axis", default=None)

    @property
    def output_limit(self) -> int:
        return 2

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self.lhs = inputs[0]
        self.rhs = inputs[1]

    def __call__(self, lhs: TileableType, rhs: TileableType):
        if self.broadcast_axis != 1 or lhs.ndim == rhs.ndim:
            self._output_types = get_output_types(lhs, rhs)
        else:
            self._output_types = [OutputType.dataframe, OutputType.dataframe]

        if lhs.ndim == rhs.ndim:
            if lhs.ndim == 1:
                return self._call_series_series(lhs, rhs)
            else:
                return self._call_dataframe_dataframe(lhs, rhs)
        else:
            if lhs.ndim == 1:
                # join order need to be reversed if not symmetric
                asym_joins = {"left", "right"} - {self.join}
                if len(asym_joins) == 1:  # self.join in {"left", "right"}
                    self.join = asym_joins.pop()
                # need to put dataframe first
                self._output_types = get_output_types(rhs, lhs)
                return self._call_dataframe_series(rhs, lhs)[::-1]
            else:
                return self._call_dataframe_series(lhs, rhs)

    def _call_dataframe_dataframe(self, lhs: TileableType, rhs: TileableType):
        l_shape = list(lhs.shape)
        r_shape = list(rhs.shape)
        if self.axis is None or self.axis == 0:
            l_idx_val = r_idx_val = self._merge_index(
                lhs.index_value, rhs.index_value, how=self.join
            )
            l_shape[0] = r_shape[0] = np.nan
        else:
            l_idx_val, r_idx_val = lhs.index_value, rhs.index_value

        if self.axis is None or self.axis == 1:
            l_empty = build_empty_df(lhs.dtypes)
            r_empty = build_empty_df(rhs.dtypes)
            aligned, _ = l_empty.align(r_empty, join=self.join, axis=self.axis)
            l_dtypes = r_dtypes = aligned.dtypes
            l_col_val = r_col_val = parse_index(aligned.columns, store_data=True)
            l_shape[1] = r_shape[1] = len(l_dtypes)
        else:
            l_dtypes, r_dtypes = lhs.dtypes, rhs.dtypes
            l_col_val, r_col_val = lhs.columns_value, rhs.columns_value

        l_kws = {
            "index_value": l_idx_val,
            "dtypes": l_dtypes,
            "shape": tuple(l_shape),
            "columns_value": l_col_val,
        }
        r_kws = {
            "index_value": r_idx_val,
            "dtypes": r_dtypes,
            "shape": tuple(r_shape),
            "columns_value": r_col_val,
        }
        return self.new_tileables([lhs, rhs], kws=[l_kws, r_kws])

    def _call_dataframe_series(self, lhs: TileableType, rhs: TileableType):
        l_shape = list(lhs.shape)
        if self.axis == 0 or self.broadcast_axis == 1:
            dtypes = lhs.dtypes
            col_val = lhs.columns_value
            l_idx_val = r_idx_val = self._merge_index(
                lhs.index_value, rhs.index_value, how=self.join
            )
            l_shape[0] = r_size = np.nan
        else:
            l_idx_val = lhs.index_value
            if not rhs.index_value.has_value():
                dtypes = None
                l_shape[1] = r_size = np.nan
                col_val = r_idx_val = self._merge_index(
                    lhs.columns_value, rhs.index_value, how=self.join
                )
            else:
                series_index = rhs.index_value.to_pandas()
                dtypes = lhs.dtypes.reindex(
                    lhs.dtypes.index.join(series_index, how=self.join)
                ).fillna(np.dtype(np.float_))
                l_shape[1] = r_size = len(dtypes)
                col_val = r_idx_val = parse_index(dtypes.index, store_data=True)

        l_kws = {
            "index_value": l_idx_val,
            "dtypes": dtypes,
            "shape": tuple(l_shape),
            "columns_value": col_val,
        }
        if self.broadcast_axis == 1:
            r_kws = {
                "index_value": r_idx_val,
                "dtypes": dtypes,
                "shape": tuple(l_shape),
                "columns_value": col_val,
            }
        else:
            r_kws = {
                "index_value": r_idx_val,
                "shape": (r_size,),
                "dtype": rhs.dtype,
            }
        return self.new_tileables([lhs, rhs], kws=[l_kws, r_kws])

    def _call_series_series(self, lhs: TileableType, rhs: TileableType):
        idx = self._merge_index(lhs.index_value, rhs.index_value, how=self.join)
        kws = [
            {"index_value": idx, "shape": (np.nan,), "dtype": lhs.dtype},
            {"index_value": idx, "shape": (np.nan,), "dtype": rhs.dtype},
        ]
        return self.new_tileables([lhs, rhs], kws=kws)

    @staticmethod
    def _merge_index(
        left_index_value: IndexValue, right_index_value: IndexValue, how: str = "outer"
    ):
        left_pd = left_index_value.to_pandas()
        right_pd = right_index_value.to_pandas()

        if not left_index_value.has_value() or not right_index_value.has_value():
            left_pd = left_pd[:0]
            right_pd = right_pd[:0]
            store_data = False
        else:
            store_data = True

        joined = left_pd.join(right_pd, how=how)
        if store_data:
            return parse_index(joined, store_data=store_data)
        else:
            return parse_index(
                joined,
                {left_index_value.key, right_index_value.key},
                store_data=store_data,
            )

    @classmethod
    def _select_nsplits(
        cls, op: "DataFrameAlign", tileable: TileableType, val_to_replace: list
    ):
        if op.axis is None:
            return val_to_replace[: tileable.ndim]
        else:
            attr_val = tileable.nsplits
            axis = op.axis % tileable.ndim
            return [
                tuple(val_to_replace[op.axis]) if i == axis else attr_val[i]
                for i in range(len(attr_val))
            ]

    @classmethod
    def _build_tiled_kw(
        cls, op: "DataFrameAlign", idx: int, chunks: list, nsplits: list
    ):
        in_tileable = op.inputs[idx]
        out_tileable = op.outputs[idx]
        kw = out_tileable.params.copy()
        kw.update(
            {
                "chunks": chunks,
                "nsplits": tuple(cls._select_nsplits(op, in_tileable, nsplits)),
            }
        )
        return kw

    @classmethod
    def _check_align_needed(
        cls, op: "DataFrameAlign", left_chunks: list, right_chunks: list
    ):
        lhs, rhs = op.lhs, op.rhs
        if all(lc.key == rc.key for lc, rc in zip(lhs.chunks, left_chunks)) and all(
            lc.key == rc.key for lc, rc in zip(rhs.chunks, right_chunks)
        ):
            raise _NoNeedToAlign

    @classmethod
    def _tile_dataframe_dataframe(cls, op: "DataFrameAlign"):
        lhs, rhs = op.lhs, op.rhs
        nsplits, chunk_shapes, left_chunks, right_chunks = align_dataframe_dataframe(
            lhs, rhs, axis=op.axis
        )
        cls._check_align_needed(op, left_chunks, right_chunks)

        left_chunk_array = np.array(left_chunks, dtype="O").reshape(chunk_shapes[0])
        right_chunk_array = np.array(right_chunks, dtype="O").reshape(chunk_shapes[1])

        left_idx_to_chunk = dict()
        l_chunks, r_chunks = [], []

        iterator = np.nditer(right_chunk_array, flags=["refs_ok", "multi_index"])
        for rc_obj in iterator:
            rc = rc_obj.tolist()
            r_index = iterator.multi_index
            l_index = tuple(r_index[i] % chunk_shapes[0][i] for i in (0, 1))
            lc = left_chunk_array[l_index]

            kws = [lc.params, rc.params]
            kws[0]["index"] = l_index
            kws[1]["index"] = r_index

            chunk_op = op.copy().reset_key()
            l_chunk, r_chunk = chunk_op.new_chunks([lc, rc], kws=kws)
            left_idx_to_chunk[l_index] = l_chunk
            r_chunks.append(r_chunk)

        iterator = np.nditer(left_chunk_array, flags=["refs_ok", "multi_index"])
        for lc_obj in iterator:
            lc = lc_obj.tolist()
            l_index = iterator.multi_index
            try:
                l_chunk = left_idx_to_chunk[l_index]
                l_chunks.append(l_chunk)
                continue
            except KeyError:
                pass

            r_index = tuple(l_index[i] % chunk_shapes[1][i] for i in (0, 1))
            rc = right_chunk_array[r_index]

            kws = [lc.params, rc.params]
            kws[0]["index"] = l_index

            chunk_op = op.copy().reset_key()
            l_chunk, _r_chunk = chunk_op.new_chunks([lc, rc], kws=kws)
            l_chunks.append(l_chunk)

        return nsplits, l_chunks, r_chunks

    @classmethod
    def _tile_dataframe_series(cls, op: "DataFrameAlign"):
        lhs, rhs = op.lhs, op.rhs
        nsplits, left_chunk_shape, left_chunks, right_chunks = align_dataframe_series(
            lhs, rhs, axis=op.axis
        )
        cls._check_align_needed(op, left_chunks, right_chunks)

        left_chunk_array = np.array(left_chunks, dtype="O").reshape(left_chunk_shape)
        axis = op.axis if op.broadcast_axis != 1 else 0
        l_chunks, r_chunks = [], []
        iterator = np.nditer(left_chunk_array, flags=["refs_ok", "multi_index"])
        for c_obj in iterator:
            c = c_obj.tolist()
            l_index = iterator.multi_index

            right_chunk = right_chunks[l_index[axis]]
            kws = [c.params, right_chunk.params]
            kws[0]["index"] = l_index
            if op.broadcast_axis != 1:
                kws[1]["index"] = (l_index[axis],)
            else:
                kws[1]["index"] = l_index

            chunk_op = op.copy().reset_key()
            l_chunk, r_chunk = chunk_op.new_chunks([c, right_chunk], kws=kws)

            l_chunks.append(l_chunk)
            if op.broadcast_axis == 1 or l_index[1 - axis] == 0:
                r_chunks.append(r_chunk)

        return nsplits, l_chunks, r_chunks

    @classmethod
    def _tile_series_series(cls, op: "DataFrameAlign"):
        nsplits, _, left_chunks, right_chunks = align_series_series(op.lhs, op.rhs)
        cls._check_align_needed(op, left_chunks, right_chunks)

        l_chunks, r_chunks = [], []
        for idx, (lc, rc) in enumerate(zip(left_chunks, right_chunks)):
            kws = [lc.params, rc.params]
            kws[0]["index"] = kws[1]["index"] = (idx,)

            chunk_op = op.copy().reset_key()
            l_chunk, r_chunk = chunk_op.new_chunks([lc, rc], kws=kws)
            l_chunks.append(l_chunk)
            r_chunks.append(r_chunk)
        return nsplits, l_chunks, r_chunks

    @classmethod
    def _tile_with_fillna(cls, tileable: TileableType):
        op = tileable.op
        if op.method is None:
            return tileable
        axis = op.fill_axis if tileable.ndim == 2 else 0
        tileable = tileable.fillna(method=op.method, limit=op.limit, axis=axis)
        return (yield from recursive_tile(tileable))

    @classmethod
    def _make_direct_output_kws(cls, left: TileableType, right: TileableType):
        kws = [left.params, right.params]
        kws[0].update(dict(chunks=left.chunks, nsplits=left.nsplits))
        kws[1].update(dict(chunks=right.chunks, nsplits=right.nsplits))
        return kws

    @classmethod
    def tile(cls, op: "DataFrameAlign"):
        try:
            if op.lhs.ndim == op.rhs.ndim:
                if op.lhs.ndim == 2:
                    nsplits, left_chunks, right_chunks = cls._tile_dataframe_dataframe(
                        op
                    )
                else:
                    nsplits, left_chunks, right_chunks = cls._tile_series_series(op)
            else:
                nsplits, left_chunks, right_chunks = cls._tile_dataframe_series(op)
        except _NoNeedToAlign:
            kws = cls._make_direct_output_kws(op.lhs, op.rhs)
        else:
            kws = [
                cls._build_tiled_kw(op, 0, left_chunks, nsplits),
                cls._build_tiled_kw(op, 1, right_chunks, nsplits),
            ]
        new_left, new_right = op.copy().new_tileables(op.inputs, kws=kws)

        new_left_filled = yield from cls._tile_with_fillna(new_left)
        new_right_filled = yield from cls._tile_with_fillna(new_right)
        if new_left_filled is not new_left or new_right_filled is not new_right:
            kws = cls._make_direct_output_kws(new_left_filled, new_right_filled)
            new_left, new_right = op.copy().new_tileables(op.inputs, kws=kws)

        return [new_left, new_right]

    @classmethod
    def execute(cls, ctx, op: "DataFrameAlign"):
        lhs_val = ctx[op.lhs.key]
        rhs_val = ctx[op.rhs.key]
        l_res, r_res = lhs_val.align(
            rhs_val,
            axis=op.axis,
            join=op.join,
            fill_value=op.fill_value,
            broadcast_axis=op.broadcast_axis,
        )
        ctx[op.outputs[0].key] = l_res
        ctx[op.outputs[1].key] = r_res


def align(
    df,
    other,
    join: str = "outer",
    axis: Union[int, str, None] = None,
    level: Union[int, str, None] = None,
    copy: bool = True,
    fill_value: Any = None,
    method: str = None,
    limit: Optional[int] = None,
    fill_axis: Union[int, str] = 0,
    broadcast_axis: Union[int, str] = None,
):
    """
    Align two objects on their axes with the specified join method.

    Join method is specified for each axis Index.

    Parameters
    ----------
    other : DataFrame or Series
    join : {'outer', 'inner', 'left', 'right'}, default 'outer'
    axis : allowed axis of the other object, default None
        Align on index (0), columns (1), or both (None).
    level : int or level name, default None
        Broadcast across a level, matching Index values on the
        passed MultiIndex level.
    copy : bool, default True
        Always returns new objects. If copy=False and no reindexing is
        required then original objects are returned.
    fill_value : scalar, default np.NaN
        Value to use for missing values. Defaults to NaN, but can be any
        "compatible" value.
    method : {'backfill', 'bfill', 'pad', 'ffill', None}, default None
        Method to use for filling holes in reindexed Series:

        - pad / ffill: propagate last valid observation forward to next valid.
        - backfill / bfill: use NEXT valid observation to fill gap.

    limit : int, default None
        If method is specified, this is the maximum number of consecutive
        NaN values to forward/backward fill. In other words, if there is
        a gap with more than this number of consecutive NaNs, it will only
        be partially filled. If method is not specified, this is the
        maximum number of entries along the entire axis where NaNs will be
        filled. Must be greater than 0 if not None.
    fill_axis : {0 or 'index', 1 or 'columns'}, default 0
        Filling axis, method and limit.
    broadcast_axis : {0 or 'index', 1 or 'columns'}, default None
        Broadcast values along this axis, if aligning two objects of
        different dimensions.

    Notes
    -----
    Currently argument `level` is not supported.

    Returns
    -------
    (left, right) : (DataFrame, type of other)
        Aligned objects.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> import mars.dataframe as md
    >>> df = md.DataFrame(
    ...     [[1, 2, 3, 4], [6, 7, 8, 9]], columns=["D", "B", "E", "A"], index=[1, 2]
    ... )
    >>> other = md.DataFrame(
    ...     [[10, 20, 30, 40], [60, 70, 80, 90], [600, 700, 800, 900]],
    ...     columns=["A", "B", "C", "D"],
    ...     index=[2, 3, 4],
    ... )
    >>> df.execute()
       D  B  E  A
    1  1  2  3  4
    2  6  7  8  9
    >>> other.execute()
        A    B    C    D
    2   10   20   30   40
    3   60   70   80   90
    4  600  700  800  900

    Align on columns:

    >>> left, right = df.align(other, join="outer", axis=1)
    >>> left.execute()
       A  B   C  D  E
    1  4  2 NaN  1  3
    2  9  7 NaN  6  8
    >>> right.execute()
        A    B    C    D   E
    2   10   20   30   40 NaN
    3   60   70   80   90 NaN
    4  600  700  800  900 NaN

    We can also align on the index:

    >>> left, right = df.align(other, join="outer", axis=0)
    >>> left.execute()
        D    B    E    A
    1  1.0  2.0  3.0  4.0
    2  6.0  7.0  8.0  9.0
    3  NaN  NaN  NaN  NaN
    4  NaN  NaN  NaN  NaN
    >>> right.execute()
        A      B      C      D
    1    NaN    NaN    NaN    NaN
    2   10.0   20.0   30.0   40.0
    3   60.0   70.0   80.0   90.0
    4  600.0  700.0  800.0  900.0

    Finally, the default `axis=None` will align on both index and columns:

    >>> left, right = df.align(other, join="outer", axis=None)
    >>> left.execute()
         A    B   C    D    E
    1  4.0  2.0 NaN  1.0  3.0
    2  9.0  7.0 NaN  6.0  8.0
    3  NaN  NaN NaN  NaN  NaN
    4  NaN  NaN NaN  NaN  NaN
    >>> right.execute()
           A      B      C      D   E
    1    NaN    NaN    NaN    NaN NaN
    2   10.0   20.0   30.0   40.0 NaN
    3   60.0   70.0   80.0   90.0 NaN
    4  600.0  700.0  800.0  900.0 NaN
    """
    axis = validate_axis(axis) if axis is not None else None
    fill_axis = validate_axis(fill_axis) if fill_axis is not None else None
    broadcast_axis = (
        validate_axis(broadcast_axis) if broadcast_axis is not None else None
    )

    if level is not None:
        raise NotImplementedError(f"Argument `level` not supported")
    if df.ndim != other.ndim and axis is None:
        raise ValueError("Must specify axis=0 or 1")

    op = DataFrameAlign(
        join=join,
        axis=axis,
        level=level,
        copy=copy,
        fill_value=fill_value,
        method=method,
        limit=limit,
        fill_axis=fill_axis,
        broadcast_axis=broadcast_axis,
    )
    return op(df, other)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import numpy as np
import pandas as pd

from ... import opcodes
from ...core import recursive_tile
from ...core.custom_log import redirect_custom_log
from ...serialization.serializables import (
    DictField,
    FunctionField,
    Int32Field,
    KeyField,
    StringField,
    TupleField,
)
from ...utils import enter_current_session, has_unknown_shape, quiet_stdio
from ..operands import DataFrameOperand, OutputType
from ..utils import (
    build_df,
    build_empty_df,
    build_series,
    parse_index,
    validate_output_types,
)
from .core import DataFrameAutoMergeMixin

logger = logging.getLogger(__name__)


class DataFrameCartesianChunk(DataFrameOperand, DataFrameAutoMergeMixin):
    _op_type_ = opcodes.CARTESIAN_CHUNK

    left = KeyField("left")
    right = KeyField("right")
    func = FunctionField("func")
    args = TupleField("args")
    kwargs = DictField("kwargs")
    auto_merge = StringField("auto_merge")
    auto_merge_threshold = Int32Field("auto_merge_threshold")

    def __init__(self, output_types=None, **kw):
        super().__init__(_output_types=output_types, **kw)
        if self.memory_scale is None:
            self.memory_scale = 2.0

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self.left = self.inputs[0]
        self.right = self.inputs[1]

    @staticmethod
    def _build_test_obj(obj):
        return (
            build_df(obj, size=2)
            if obj.ndim == 2
            else build_series(obj, size=2, name=obj.name)
        )

    def __call__(self, left, right, index=None, dtypes=None):
        test_left = self._build_test_obj(left)
        test_right = self._build_test_obj(right)
        output_type = self.output_types[0] if self.output_types else None

        if output_type == OutputType.df_or_series:
            return self.new_df_or_series([left, right])

        # try run to infer meta
        try:
            with np.errstate(all="ignore"), quiet_stdio():
                obj = self.func(test_left, test_right, *self.args, **self.kwargs)
        except:  # noqa: E722  # nosec  # pylint: disable=bare-except
            if output_type == OutputType.series:
                obj = pd.Series([], dtype=np.dtype(object))
            elif output_type == OutputType.dataframe and dtypes is not None:
                obj = build_empty_df(dtypes)
            else:
                raise TypeError(
                    "Cannot determine `output_type`, "
                    "you have to specify it as `dataframe` or `series`, "
                    "for dataframe, `dtypes` is required as well "
                    "if output_type='dataframe'"
                )

        if getattr(obj, "ndim", 0) == 1 or output_type == OutputType.series:
            shape = self.kwargs.pop("shape", (np.nan,))
            if index is None:
                index = obj.index
            index_value = parse_index(
                index, left, right, self.func, self.args, self.kwargs
            )
            return self.new_series(
                [left, right],
                dtype=obj.dtype,
                shape=shape,
                index_value=index_value,
                name=obj.name,
            )
        else:
            dtypes = dtypes if dtypes is not None else obj.dtypes
            # dataframe
            shape = (np.nan, len(dtypes))
            columns_value = parse_index(dtypes.index, store_data=True)
            if index is None:
                index = obj.index
            index_value = parse_index(
                index, left, right, self.func, self.args, self.kwargs
            )
            return self.new_dataframe(
                [left, right],
                shape=shape,
                dtypes=dtypes,
                index_value=index_value,
                columns_value=columns_value,
            )

    @classmethod
    def tile(cls, op: "DataFrameCartesianChunk"):
        left = op.left
        right = op.right
        out = op.outputs[0]
        out_type = op.output_types[0]

        auto_merge_threshold = op.auto_merge_threshold
        auto_merge_before, auto_merge_after = cls._get_auto_merge_options(op.auto_merge)

        merge_before_res = yield from cls._merge_before(
            op, auto_merge_before, auto_merge_threshold, left, right, logger
        )
        left, right = merge_before_res[0], merge_before_res[1]

        if left.ndim == 2 and left.chunk_shape[1] > 1:
            if has_unknown_shape(left):
                yield
            # if left is a DataFrame, make sure 1 chunk on axis columns
            left = yield from recursive_tile(left.rechunk({1: left.shape[1]}))
        if right.ndim == 2 and right.chunk_shape[1] > 1:
            if has_unknown_shape(right):
                yield
            # if right is a DataFrame, make sure 1 chunk on axis columns
            right = yield from recursive_tile(right.rechunk({1: right.shape[1]}))

        out_chunks = []
        if out_type == OutputType.dataframe:
            nsplits = [[], [out.shape[1]]]
        elif out_type == OutputType.series:
            nsplits = [[]]
        else:
            # DataFrameOrSeries
            nsplits = None
        i = 0
        for left_chunk in left.chunks:
            for right_chunk in right.chunks:
                chunk_op = op.copy().reset_key()
                chunk_op.tileable_op_key = op.key
                if out_type == OutputType.df_or_series:
                    out_chunks.append(
                        chunk_op.new_chunk(
                            [left_chunk, right_chunk], index=(i, 0), collapse_axis=1
                        )
                    )
                elif out_type == OutputType.dataframe:
                    shape = (np.nan, out.shape[1])
                    index_value = parse_index(
                        out.index_value.to_pandas(),
                        left_chunk,
                        right_chunk,
                        op.func,
                        op.args,
                        op.kwargs,
                    )
                    out_chunk = chunk_op.new_chunk(
                        [left_chunk, right_chunk],
                        shape=shape,
                        index_value=index_value,
                        columns_value=out.columns_value,
                        dtypes=out.dtypes,
                        index=(i, 0),
                    )
                    out_chunks.append(out_chunk)
                    nsplits[0].append(out_chunk.shape[0])
                else:
                    shape = (np.nan,)
                    index_value = parse_index(
                        out.index_value.to_pandas(),
                        left_chunk,
                        right_chunk,
                        op.func,
                        op.args,
                        op.kwargs,
                    )
                    out_chunk = chunk_op.new_chunk(
                        [left_chunk, right_chunk],
                        shape=shape,
                        index_value=index_value,
                        dtype=out.dtype,
                        name=out.name,
                        index=(i,),
                    )
                    out_chunks.append(out_chunk)
                    nsplits[0].append(out_chunk.shape[0])
                i += 1

        params = out.params
        params["nsplits"] = tuple(tuple(ns) for ns in nsplits) if nsplits else nsplits
        params["chunks"] = out_chunks
        new_op = op.copy()
        ret = new_op.new_tileables(op.inputs, kws=[params])

        ret = yield from cls._merge_after(
            op, auto_merge_after, auto_merge_threshold, ret, logger
        )
        return ret

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx, op: "DataFrameCartesianChunk"):
        left, right = ctx[op.left.key], ctx[op.right.key]
        ctx[op.outputs[0].key] = op.func(left, right, *op.args, **(op.kwargs or dict()))


def cartesian_chunk(
    left,
    right,
    func,
    skip_infer=False,
    args=(),
    auto_merge: str = "both",
    auto_merge_threshold: int = 8,
    **kwargs,
):
    output_type = kwargs.pop("output_type", None)
    output_types = kwargs.pop("output_types", None)
    object_type = kwargs.pop("object_type", None)
    output_types = validate_output_types(
        output_type=output_type, output_types=output_types, object_type=object_type
    )
    output_type = output_types[0] if output_types else None
    if output_type:
        output_types = [output_type]
    elif skip_infer:
        output_types = [OutputType.df_or_series]
    index = kwargs.pop("index", None)
    dtypes = kwargs.pop("dtypes", None)
    memory_scale = kwargs.pop("memory_scale", None)
    if auto_merge not in ["both", "none", "before", "after"]:  # pragma: no cover
        raise ValueError(
            f"auto_merge can only be `both`, `none`, `before` or `after`, got {auto_merge}"
        )

    op = DataFrameCartesianChunk(
        left=left,
        right=right,
        func=func,
        args=args,
        kwargs=kwargs,
        output_types=output_types,
        memory_scale=memory_scale,
        auto_merge=auto_merge,
        auto_merge_threshold=auto_merge_threshold,
    )
    return op(left, right, index=index, dtypes=dtypes)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

import pandas as pd

try:
    from pandas.core.arraylike import OpsMixin as PdOpsMixin
except ImportError:  # pragma: no cover
    PdOpsMixin = None

from ..core import DATAFRAME_TYPE, INDEX_TYPE, SERIES_TYPE, is_build_mode
from ..ufunc.tensor import register_tensor_ufunc
from ..utils import wrap_notimplemented_exception
from .abs import DataFrameAbs, abs_
from .add import DataFrameAdd, add, radd
from .arccos import DataFrameArccos
from .arccosh import DataFrameArccosh
from .arcsin import DataFrameArcsin
from .arcsinh import DataFrameArcsinh
from .arctan import DataFrameArctan
from .arctanh import DataFrameArctanh
from .around import DataFrameAround, around
from .bitwise_and import DataFrameAnd, bitand, rbitand
from .bitwise_or import DataFrameOr, bitor, rbitor
from .bitwise_xor import DataFrameXor, bitxor, rbitxor
from .ceil import DataFrameCeil
from .cos import DataFrameCos
from .cosh import DataFrameCosh
from .degrees import DataFrameDegrees
from .dot import dot, rdot
from .equal import DataFrameEqual, eq
from .exp import DataFrameExp
from .exp2 import DataFrameExp2
from .expm1 import DataFrameExpm1
from .floor import DataFrameFloor
from .floordiv import DataFrameFloorDiv, floordiv, rfloordiv
from .greater import DataFrameGreater, gt
from .greater_equal import DataFrameGreaterEqual, ge
from .invert import DataFrameNot, invert
from .is_ufuncs import DataFrameIsFinite, DataFrameIsInf, DataFrameIsNan
from .less import DataFrameLess, lt
from .less_equal import DataFrameLessEqual, le
from .log import DataFrameLog
from .log2 import DataFrameLog2
from .log10 import DataFrameLog10
from .mod import DataFrameMod, mod, rmod
from .multiply import DataFrameMul, mul, rmul
from .negative import DataFrameNegative, negative
from .not_equal import DataFrameNotEqual, ne
from .power import DataFramePower, power, rpower
from .radians import DataFrameRadians
from .sin import DataFrameSin
from .sinh import DataFrameSinh
from .sqrt import DataFrameSqrt
from .subtract import DataFrameSubtract, rsubtract, subtract
from .tan import DataFrameTan
from .tanh import DataFrameTanh
from .truediv import DataFrameTrueDiv, rtruediv, truediv


def _wrap_eq():
    @functools.wraps(eq)
    def call(df, other, **kw):
        if is_build_mode():
            return df._equals(other)
        return _wrap_comparison(eq)(df, other, **kw)

    return call


def _wrap_comparison(func):
    @functools.wraps(func)
    def call(df, other, **kw):
        if isinstance(df, DATAFRAME_TYPE) and isinstance(other, DATAFRAME_TYPE):
            # index and columns should be identical
            for index_type in ["index_value", "columns_value"]:
                left, right = getattr(df, index_type), getattr(other, index_type)
                if left.has_value() and right.has_value():
                    # if df and other's index or columns has value
                    index_eq = left.to_pandas().equals(right.to_pandas())
                else:
                    index_eq = left.key == right.key
                if not index_eq:
                    raise ValueError(
                        "Can only compare identically-labeled DataFrame object"
                    )
        return wrap_notimplemented_exception(func)(df, other, **kw)

    return call


_reverse_magic_names = {
    "eq": "eq",
    "ne": "ne",
    "lt": "ge",
    "le": "gt",
    "gt": "le",
    "ge": "lt",
}


def _wrap_pandas_magics(cls, magic_name: str):
    magic_func_name = f"__{magic_name}__"
    magic_rfunc_name = _reverse_magic_names.get(magic_name, f"__r{magic_name}__")
    try:
        raw_method = getattr(cls, magic_func_name)
    except AttributeError:
        return

    @functools.wraps(raw_method)
    def wrapped(self, other):
        if not isinstance(other, (DATAFRAME_TYPE, SERIES_TYPE, INDEX_TYPE)):
            return raw_method(self, other)

        try:
            val = getattr(other, magic_rfunc_name)(self)
        except AttributeError:  # pragma: no cover
            return raw_method(self, other)

        if val is NotImplemented:  # pragma: no cover
            return raw_method(self, other)
        return val

    setattr(cls, magic_func_name, wrapped)


def _install():
    def _register_method(cls, name, func, wrapper=None):
        if wrapper is None:

            @functools.wraps(func)
            def wrapper(df, *args, **kwargs):
                return func(df, *args, **kwargs)

        try:
            if issubclass(cls, DATAFRAME_TYPE):
                wrapper.__doc__ = func.__frame_doc__
            elif issubclass(cls, SERIES_TYPE):
                wrapper.__doc__ = func.__series_doc__
            else:
                wrapper = func
        except AttributeError:
            wrapper = func

        wrapper.__name__ = func.__name__
        setattr(cls, name, wrapper)

    def _register_bin_method(cls, name, func):
        def call_df_fill(df, other, axis="columns", level=None, fill_value=None):
            return func(df, other, axis=axis, level=level, fill_value=fill_value)

        def call_df_no_fill(df, other, axis="columns", level=None):
            return func(df, other, axis=axis, level=level)

        def call_series_fill(df, other, level=None, fill_value=None, axis=0):
            return func(df, other, axis=axis, level=level, fill_value=fill_value)

        def call_series_no_fill(df, other, level=None, axis=0):
            return func(df, other, axis=axis, level=level)

        if issubclass(cls, DATAFRAME_TYPE):
            call = (
                call_df_fill
                if "fill_value" in func.__code__.co_varnames
                else call_df_no_fill
            )
        elif issubclass(cls, SERIES_TYPE):
            call = (
                call_series_fill
                if "fill_value" in func.__code__.co_varnames
                else call_series_no_fill
            )
        else:
            call = None
        return _register_method(cls, name, func, wrapper=call)

    # register mars tensor ufuncs
    ufunc_ops = [
        # unary
        DataFrameAbs,
        DataFrameLog,
        DataFrameLog2,
        DataFrameLog10,
        DataFrameSin,
        DataFrameCos,
        DataFrameTan,
        DataFrameSinh,
        DataFrameCosh,
        DataFrameTanh,
        DataFrameArcsin,
        DataFrameArccos,
        DataFrameArctan,
        DataFrameArcsinh,
        DataFrameArccosh,
        DataFrameArctanh,
        DataFrameRadians,
        DataFrameDegrees,
        DataFrameCeil,
        DataFrameFloor,
        DataFrameAround,
        DataFrameExp,
        DataFrameExp2,
        DataFrameExpm1,
        DataFrameSqrt,
        DataFrameNot,
        DataFrameIsNan,
        DataFrameIsInf,
        DataFrameIsFinite,
        DataFrameNegative,
        # binary
        DataFrameAdd,
        DataFrameEqual,
        DataFrameFloorDiv,
        DataFrameGreater,
        DataFrameGreaterEqual,
        DataFrameLess,
        DataFrameLessEqual,
        DataFrameAnd,
        DataFrameOr,
        DataFrameXor,
        DataFrameMod,
        DataFrameMul,
        DataFrameNotEqual,
        DataFramePower,
        DataFrameSubtract,
        DataFrameTrueDiv,
    ]
    for ufunc_op in ufunc_ops:
        register_tensor_ufunc(ufunc_op)

    for entity in DATAFRAME_TYPE + SERIES_TYPE:
        setattr(entity, "__abs__", abs_)
        setattr(entity, "abs", abs_)
        _register_method(entity, "round", around)
        setattr(entity, "__invert__", invert)

        setattr(entity, "__add__", wrap_notimplemented_exception(add))
        setattr(entity, "__radd__", wrap_notimplemented_exception(radd))
        _register_bin_method(entity, "add", add)
        _register_bin_method(entity, "radd", radd)

        setattr(entity, "__sub__", wrap_notimplemented_exception(subtract))
        setattr(entity, "__rsub__", wrap_notimplemented_exception(rsubtract))
        _register_bin_method(entity, "sub", subtract)
        _register_bin_method(entity, "rsub", rsubtract)

        setattr(entity, "__mul__", wrap_notimplemented_exception(mul))
        setattr(entity, "__rmul__", wrap_notimplemented_exception(rmul))
        _register_bin_method(entity, "mul", mul)
        _register_bin_method(entity, "multiply", mul)
        _register_bin_method(entity, "rmul", rmul)

        setattr(entity, "__floordiv__", wrap_notimplemented_exception(floordiv))
        setattr(entity, "__rfloordiv__", wrap_notimplemented_exception(rfloordiv))
        setattr(entity, "__truediv__", wrap_notimplemented_exception(truediv))
        setattr(entity, "__rtruediv__", wrap_notimplemented_exception(rtruediv))
        setattr(entity, "__div__", wrap_notimplemented_exception(truediv))
        setattr(entity, "__rdiv__", wrap_notimplemented_exception(rtruediv))
        _register_bin_method(entity, "floordiv", floordiv)
        _register_bin_method(entity, "rfloordiv", rfloordiv)
        _register_bin_method(entity, "truediv", truediv)
        _register_bin_method(entity, "rtruediv", rtruediv)
        _register_bin_method(entity, "div", truediv)
        _register_bin_method(entity, "rdiv", rtruediv)

        setattr(entity, "__mod__", wrap_notimplemented_exception(mod))
        setattr(entity, "__rmod__", wrap_notimplemented_exception(rmod))
        _register_bin_method(entity, "mod", mod)
        _register_bin_method(entity, "rmod", rmod)

        setattr(entity, "__pow__", wrap_notimplemented_exception(power))
        setattr(entity, "__rpow__", wrap_notimplemented_exception(rpower))
        _register_bin_method(entity, "pow", power)
        _register_bin_method(entity, "rpow", rpower)

        setattr(entity, "__eq__", _wrap_eq())
        setattr(entity, "__ne__", _wrap_comparison(ne))
        setattr(entity, "__lt__", _wrap_comparison(lt))
        setattr(entity, "__gt__", _wrap_comparison(gt))
        setattr(entity, "__ge__", _wrap_comparison(ge))
        setattr(entity, "__le__", _wrap_comparison(le))
        _register_bin_method(entity, "eq", eq)
        _register_bin_method(entity, "ne", ne)
        _register_bin_method(entity, "lt", lt)
        _register_bin_method(entity, "gt", gt)
        _register_bin_method(entity, "ge", ge)
        _register_bin_method(entity, "le", le)

        setattr(entity, "__matmul__", dot)
        setattr(entity, "__rmatmul__", rdot)
        _register_method(entity, "dot", dot)

        setattr(entity, "__and__", wrap_notimplemented_exception(bitand))
        setattr(entity, "__rand__", wrap_notimplemented_exception(rbitand))

        setattr(entity, "__or__", wrap_notimplemented_exception(bitor))
        setattr(entity, "__ror__", wrap_notimplemented_exception(rbitor))

        setattr(entity, "__xor__", wrap_notimplemented_exception(bitxor))
        setattr(entity, "__rxor__", wrap_notimplemented_exception(rbitxor))

        setattr(entity, "__neg__", wrap_notimplemented_exception(negative))

    for entity in INDEX_TYPE:
        setattr(entity, "__eq__", _wrap_eq())

    if PdOpsMixin is not None and not hasattr(
        pd, "_mars_df_arith_wrapped"
    ):  # pragma: no branch
        # wrap pandas magic functions to intercept reverse operands
        for magic_name in [
            "add",
            "sub",
            "mul",
            "div",
            "truediv",
            "floordiv",
            "mod",
            "pow",
            "and",
            "or",
            "xor",
            "eq",
            "ne",
            "lt",
            "le",
            "gt",
            "ge",
        ]:
            _wrap_pandas_magics(PdOpsMixin, magic_name)

        for pd_cls in (pd.DataFrame, pd.Series):
            _wrap_pandas_magics(pd_cls, "matmul")

        pd._mars_df_arith_wrapped = True


_install()
del _install



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType
from .core import (
    DATAFRAME_TYPE,
    DataFrameReductionMixin,
    DataFrameReductionOperand,
    recursive_tile,
)


class DataFrameAll(DataFrameReductionOperand, DataFrameReductionMixin):
    _op_type_ = OperandDef.ALL
    _func_name = "all"

    @property
    def is_atomic(self):
        return True

    @classmethod
    def tile(cls, op):
        in_df = op.inputs[0]
        out_df = op.outputs[0]
        if op.axis is None and isinstance(in_df, DATAFRAME_TYPE):
            dtypes = pd.Series([out_df.dtype])
            index = in_df.dtypes.index
            out_df = yield from recursive_tile(
                in_df.agg(
                    cls.get_reduction_callable(op),
                    axis=0,
                    _numeric_only=op.numeric_only,
                    _bool_only=op.bool_only,
                    _combine_size=op.combine_size,
                    _output_type=OutputType.series,
                    _dtypes=dtypes,
                    _index=index,
                )
            )
            out_df = yield from recursive_tile(
                out_df.agg(
                    cls.get_reduction_callable(op),
                    axis=0,
                    _numeric_only=op.numeric_only,
                    _bool_only=op.bool_only,
                    _combine_size=op.combine_size,
                    _output_type=OutputType.scalar,
                    _dtypes=out_df.dtype,
                    _index=None,
                )
            )
            return [out_df]
        else:
            return (yield from super().tile(op))

    def __call__(self, df):
        if self.axis is None and isinstance(df, DATAFRAME_TYPE):
            return self.new_scalar([df], np.dtype("bool"))
        else:
            return super().__call__(df)


def all_series(
    series,
    axis=0,
    bool_only=None,
    skipna=True,
    level=None,
    combine_size=None,
    method=None,
):
    op = DataFrameAll(
        axis=axis,
        skipna=skipna,
        level=level,
        bool_only=bool_only,
        combine_size=combine_size,
        output_types=[OutputType.scalar],
        method=method,
    )
    return op(series)


def all_dataframe(
    df,
    axis=0,
    bool_only=None,
    skipna=True,
    level=None,
    combine_size=None,
    method=None,
):
    output_types = [OutputType.series] if axis is not None else [OutputType.scalar]
    op = DataFrameAll(
        axis=axis,
        skipna=skipna,
        level=level,
        bool_only=bool_only,
        combine_size=combine_size,
        output_types=output_types,
        method=method,
    )
    return op(df)


def all_index(idx):
    op = DataFrameAll(output_types=[OutputType.scalar])
    return op(idx)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Dict

from .._mars.utils import lazy_import
from ..core.adapter import (
    collect_cls_members,
    register_data_members,
    wrap_mars_callable,
)
from ..core.data import DataType
from .backends.huggingface.from_huggingface import from_huggingface
from .dataset import Dataset

MARS_DATASET_TYPE = (Dataset,)


def _collect_module_callables() -> Dict[str, Callable]:
    module_callables = {}
    # TODO(fyrestone): Remove this hard code module
    hf_datasets = lazy_import("datasets")

    # install module functions
    for func in [from_huggingface]:
        module_callables[func.__name__] = wrap_mars_callable(
            func,
            attach_docstring=True,
            is_cls_member=False,
            docstring_src_module=hf_datasets,
            docstring_src=func,
        )
    return module_callables


MARS_DATASET_CALLABLES: Dict[str, Callable] = _collect_module_callables()


def _install():
    for cls in MARS_DATASET_TYPE:
        register_data_members(
            DataType.dataset, collect_cls_members(cls, DataType.dataset)
        )



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sklearn.preprocessing as sk_preproc

from ...._mars.learn import preprocessing as mars_preproc
from ...._mars.learn.preprocessing import LabelBinarizer as MarsLabelBinarizer
from ...._mars.learn.preprocessing import LabelEncoder as MarsLabelEncoder
from ...._mars.learn.preprocessing import MinMaxScaler as MarsMinMaxScaler
from ....core.utils.docstring import attach_module_callable_docstring
from ...utils import SKLearnBase, _collect_module_callables, _install_cls_members


class MinMaxScaler(SKLearnBase):
    _marscls = MarsMinMaxScaler


class LabelBinarizer(SKLearnBase):
    _marscls = MarsLabelBinarizer


class LabelEncoder(SKLearnBase):
    _marscls = MarsLabelEncoder


SKLEARN_PREPROC_CLS_MAP = {
    MinMaxScaler: MarsMinMaxScaler,
    LabelEncoder: MarsLabelEncoder,
    LabelBinarizer: MarsLabelBinarizer,
}

MARS_SKLEARN_PREPROC_CALLABLES = _collect_module_callables(
    mars_preproc, sk_preproc, skip_members=["register_op"]
)
_install_cls_members(
    SKLEARN_PREPROC_CLS_MAP, MARS_SKLEARN_PREPROC_CALLABLES, sk_preproc
)
attach_module_callable_docstring(MinMaxScaler, sk_preproc, sk_preproc.MinMaxScaler)
attach_module_callable_docstring(LabelBinarizer, sk_preproc, sk_preproc.LabelBinarizer)
attach_module_callable_docstring(LabelEncoder, sk_preproc, sk_preproc.LabelEncoder)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import atexit
import logging
import os
import re
import subprocess
import time

# Configure logging
logger = logging.getLogger(__name__)


class SLURMCluster:
    def __init__(
        self,
        job_name=None,
        num_nodes=None,
        partition_option=None,
        load_env=None,
        output_path=None,
        error_path=None,
        work_dir=None,
        time=None,
        processes=None,
        cores=None,
        memory=None,
        account=None,
        webport=16379,
        **kwargs,
    ):
        """
        The entrance of deploying a SLURM cluster.

        Parameters
        ----------
        job_name : str, optional
            Name of the Slurm job, by default None
        num_nodes : int, optional
            Number of nodes in the Slurm cluster, by default None
        partition_option : str, optional
            Request a specific partition for the resource allocation, by default None
        load_env : str, optional
            Conda Environment to load, by default None
        output_path : str, optional
            Path for Log output, by default None
        error_path : str, optional
            Path for Log error, by default None
        work_dir : str, optional
            Slurm‘s Working directory,the default place to receive the logs and result, by default None
        time : str, optional
            Minimum time limit on the job allocation, by default None
        processes : int, optional
            Number of processes, by default None
        cores : int, optional
            Number of cores, by default None
        memory : str, optional
            Specify the real memory required per node. Default units are megabytes, by default None
        account : str, optional
            Charge resources used by this job to specified account, by default None
        webport : int, optional
            Xorbits' Web port, by default 16379
        If user have some specifics needing for can just follow the slurm interface we add it at the end automatically
        """
        commands = ["#!/bin/bash"]

        self.job_name = job_name
        self.num_nodes = num_nodes
        self.partition_option = partition_option
        self.output_path = output_path
        self.work_dir = work_dir
        self.time = time
        self.processes = processes
        self.cores = cores
        self.memory = memory
        self.load_env = load_env
        self.error_path = error_path
        self.account = account
        slurm_params = {
            "J": self.job_name,
            "nodes": self.num_nodes,
            "partition": self.partition_option,
            "error": self.error_path,
            "output": self.output_path,
            "chdir": self.work_dir,
            "time": self.time,
            "ntasks": self.processes,
            "cpus-per-task": self.cores,
            "mem": self.memory,
            "A": self.account,
            **kwargs,
        }
        self.commands = None
        self.web_port = webport
        for param, value in slurm_params.items():
            if value is not None:
                # there are two modes of sbatch, one is like --time, the other one is like -A，so i just judge it by using len
                if len(str(param)) > 1:
                    commands.append(f"#SBATCH --{param}={value}")
                else:
                    commands.append(f"#SBATCH -{param} {value}")

        if self.load_env:
            commands.append(f"source activate {self.load_env}")

        commands += [
            "set -x",
            'nodes=$(scontrol show hostnames "$SLURM_JOB_NODELIST")',
            "nodes_array=($nodes)",
            "head_node=${nodes_array[0]}",
            "port=16380",
            f"web_port={self.web_port}",
            'echo "Starting SUPERVISOR at ${head_node}"',
            'srun --nodes=1 --ntasks=1 -w "${head_node}" \\',
            '    xorbits-supervisor -H "${head_node}" -p "${port}" -w "${web_port}"&',
            "sleep 30",
            "worker_num=$((SLURM_JOB_NUM_NODES - 1))",
            "for ((i = 1; i <= worker_num; i++)); do",
            "    node_i=${nodes_array[$i]}",
            "    port_i=$((port + i))",
            '    echo "Starting WORKER $i at ${node_i}"',
            '    srun --nodes=1 --ntasks=1 -w "${node_i}" \\',
            '        xorbits-worker -H "${node_i}"  -p "${port_i}" -s "${head_node}":"${port}"&',
            "done",
            "sleep 300",
            'address=http://"${head_node}":"${web_port}"',
        ]
        # here I give a very long sleep time to avoid when supervisor nodes don't start, and the other node can't find the supervisor node
        self.commands = "\n".join(commands)
        self.sbatch_out = ""

    def run(self):
        shell_commands = self.commands
        with open("slurm.sh", "w") as f:
            f.write(shell_commands)

        os.chmod("slurm.sh", 0o770)

        result = subprocess.run(["sbatch", "slurm.sh"], capture_output=True, text=True)

        if result.returncode == 0:
            logger.info("Job submitted successfully.")
            self.sbatch_out = result.stdout
            self.job_id = self.get_job_id()
            if self.job_id:
                logger.info(f"Job ID is {self.job_id}.")
                atexit.register(self.cancel_job)
            else:
                logger.error("Could not get job ID. Cleanup upon exit is not possible.")
            return self.get_job_address()
        else:
            logger.error("Job submission failed.")
            logger.error("Output: {}".format(result.stdout))
            logger.error("Errors: {}".format(result.stderr))

            return None

    def get_job_id(self):
        sbatch_output = self.sbatch_out
        job_id = None
        for line in sbatch_output.split("\n"):
            if "Submitted batch job" in line:
                job_id = line.split(" ")[-1]
        return job_id

    def get_sbatch_out(self):
        logging.info(f"getting batch_out:{self.sbatch_out}")
        return self.sbatch_out

    def cancel_job(self):
        if self.job_id:
            logger.info(f"Cancelling job {self.job_id}")
            subprocess.run(["scancel", self.job_id])

    def update_head_node(self):
        try:
            if self.job_id:
                time.sleep(5)
                command = ["scontrol", "show", "job", self.job_id]
                result = subprocess.run(command, capture_output=True, text=True)
                node_list = None
                if result.returncode == 0:
                    job_info = result.stdout
                    node_list_pattern = r"NodeList=(c\[\d+-\d+\]|c\d)"
                    matches = re.search(node_list_pattern, job_info)

                if matches:
                    node_list = matches.group(1)
                    logger.info(f"NodeList:{node_list}")
                    if node_list is None:
                        raise ValueError(
                            f"Job {self.job_id} not found or NodeList information not available."
                        )
                    # get_head_node from nodelist
                    if "[" in node_list:
                        head_node = node_list.split("-")[0].replace("[", "")
                    else:
                        # when only one node
                        head_node = node_list

                    self.head_node = head_node
                    return head_node
                else:
                    logger.warning("NodeList not found in the string.")

        except subprocess.CalledProcessError as e:
            logger.error(f"Error executing scontrol: {e}")
        except ValueError as e:
            logger.error(f"Error: {e}")
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")

        self.head_node = None
        logger.warning("Failed to retrieve head node.")

    def get_job_address(self, retry_attempts=10, sleep_interval=30):
        # We retry several times to get job data
        for attempt in range(retry_attempts):
            try:
                self.update_head_node()
                if self.head_node is not None:
                    address = f"http://{self.head_node}:{self.web_port}"
                    return address
                else:
                    logger.warning(
                        f"Attempt {attempt + 1} failed, retrying after {sleep_interval}s..."
                    )
                    time.sleep(sleep_interval)
            except Exception as e:
                logger.error(str(e))



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

import numpy
import pandas

from ...core.data import DataType

_DATA_TYPE_TO_DOCSTRING_SRC: Dict[DataType, Tuple[ModuleType, Type]] = {
    DataType.dataframe: (pandas, pandas.DataFrame),
    DataType.series: (pandas, pandas.Series),
    DataType.index: (pandas, pandas.Index),
    DataType.categorical: (pandas, pandas.Categorical),
    DataType.dataframe_groupby: (
        pandas,
        pandas.core.groupby.generic.DataFrameGroupBy,
    ),
    DataType.series_groupby: (pandas, pandas.core.groupby.generic.SeriesGroupBy),
    DataType.tensor: (numpy, numpy.ndarray),
}


def get_base_indentation(doc: str) -> Optional[str]:
    # the layout of a pandas docstring is:
    # \n<base indentation><line>\n\n<base indentation><line>...
    # the layout of a numpy docstring is:
    # <line>\n\n<base indentation><line>...
    # thus, we need to find a line that starts with '\n\n' and the number of its leading
    # spaces would be the base indentation.
    idx = doc.find("\n\n")
    if idx + 2 < len(doc):
        doc = doc[idx + 2 :]
        return " " * (len(doc) - len(doc.lstrip(" ")))
    else:  # pragma: no cover
        return None


def add_docstring_disclaimer(
    docstring_src_module: Optional[ModuleType],
    docstring_src_cls: Optional[Callable],
    doc: Optional[str],
    fallback_warning: bool = False,
) -> Optional[str]:
    if doc is None:
        return None

    base_indentation = get_base_indentation(doc)
    if base_indentation is None:  # pragma: no cover
        return doc

    warning_msg = (
        f"\n\n{base_indentation}.. warning:: This method has not been implemented yet. Xorbits will try to "
        f"execute it with {docstring_src_module.__name__}."
        if fallback_warning and docstring_src_module
        else "\n"
    )

    if (
        docstring_src_cls is not None
        and hasattr(docstring_src_cls, "__module__")
        and docstring_src_cls.__module__
    ):
        return (
            doc + warning_msg + f"\n{base_indentation}This docstring was copied from "
            f"{docstring_src_cls.__module__}.{docstring_src_cls.__name__}."
        )
    elif docstring_src_module is not None:
        return (
            doc
            + warning_msg
            + f"\n{base_indentation}This docstring was copied from {docstring_src_module.__name__}."
        )
    else:
        return doc


def add_version_disclaimer(
    doc: Optional[str], docstring_src_module: Optional[ModuleType]
) -> Optional[str]:
    if doc is None or docstring_src_module is None:
        return doc

    lines = []
    for line in doc.splitlines():
        stripped = line.strip()
        if (
            stripped.startswith(".. deprecated::")
            or stripped.startswith(".. versionchanged::")
            or stripped.startswith(".. versionadded::")
        ):
            line = line + f"({docstring_src_module.__name__})"
        lines.append(line)

    return "\n".join(lines)


def skip_doctest(doc: Optional[str]) -> Optional[str]:
    def skip_line(line):
        # NumPy docstring contains cursor and comment only example
        stripped = line.strip()
        if stripped == ">>>" or stripped.startswith(">>> #"):
            return line
        elif ">>>" in stripped and "+SKIP" not in stripped:
            if "# doctest:" in line:
                return line + ", +SKIP"
            else:
                return line + "  # doctest: +SKIP"
        else:
            return line

    if doc is None:
        return None
    return "\n".join([skip_line(line) for line in doc.splitlines()])


def add_arg_disclaimer(
    src: Optional[Any],
    dest: Optional[Any],
    doc: Optional[str],
) -> Optional[str]:
    import re

    def get_named_args(func: Callable) -> List[str]:
        s = inspect.signature(func)
        return [
            n
            for n, p in s.parameters.items()
            if p.kind in [p.POSITIONAL_OR_KEYWORD, p.POSITIONAL_ONLY, p.KEYWORD_ONLY]
        ]

    def mark_unsupported_args(doc: str, args: List[str]) -> str:
        import re

        lines = doc.split("\n")
        for arg in args:
            subset = [
                (i, line)
                for i, line in enumerate(lines)
                if re.match(r"^\s*" + arg + " ?:", line)
            ]
            if len(subset) == 1:
                [(i, line)] = subset
                lines[i] = line + "  (Not supported yet)"
        return "\n".join(lines)

    def add_extra_args(doc: str, original_doc: Optional[Any], args: List[str]) -> str:
        def count_leading_spaces(s: str):
            return len(s) - len(s.lstrip(" "))

        def get_param_name(s: str) -> Optional[str]:
            # example:
            # '    data : ndarray (structured or homogeneous), Iterable, dict, or DataFrame'
            m = re.match(r"\s*(\S+)\s*:.*", s)
            if m:
                return m.group(1)
            return None

        def is_param_description(s: str, base_indentation: str) -> bool:
            if s.strip():
                return count_leading_spaces(s) == len(base_indentation) + 4
            else:
                return True

        if original_doc is None:
            return doc

        # the indentation of generated docstring must be consistent.
        base_indentation = get_base_indentation(doc)
        original_base_indentation = get_base_indentation(original_doc)
        if (
            base_indentation is None or original_base_indentation is None
        ):  # pragma: no cover
            return doc
        param_description_indentation = base_indentation + " " * 4

        new_section: List[str] = []
        lines = original_doc.splitlines()
        idx = 0
        parameter_pattern = re.compile(r"\s+Parameters")
        while idx < len(lines):
            if re.match(parameter_pattern, lines[idx]):
                idx += 2
                while idx < len(lines) and get_param_name(lines[idx]) is not None:
                    param_name = get_param_name(lines[idx])
                    if param_name in args:
                        if not new_section:
                            # init new section only if the original docstring has description of
                            # any extra parameter.
                            new_section.append("")
                            new_section.append(base_indentation + "Extra Parameters")
                            new_section.append(
                                base_indentation + "-" * len("Extra Parameters")
                            )
                        new_section.append(base_indentation + lines[idx].strip())
                        idx += 1
                        while idx < len(lines) and is_param_description(
                            lines[idx], original_base_indentation
                        ):
                            if lines[idx].strip():
                                new_section.append(
                                    param_description_indentation + lines[idx].strip()
                                )
                            idx += 1
                    else:
                        # skip the parameter description.
                        idx += 1
                        while idx < len(lines) and is_param_description(
                            lines[idx], original_base_indentation
                        ):
                            idx += 1
                new_section.append(base_indentation)
                break
            idx += 1

        return doc + "\n".join(new_section)

    if doc is None:
        return None
    if src is None or not callable(src) or dest is None or not callable(dest):
        return doc

    try:
        src_args = get_named_args(src)
        dest_args = get_named_args(dest)
        unsupported_args = [a for a in src_args if a not in dest_args]
        extra_args = [a for a in dest_args if a not in src_args]
        if unsupported_args:
            doc = mark_unsupported_args(doc, unsupported_args)
        if extra_args:
            doc = add_extra_args(doc, getattr(dest, "__doc__", None), extra_args)
    except ValueError:
        return doc

    return doc


def gen_member_docstring(
    docstring_src_cls: Optional[Type],
    member_name: str,
) -> Optional[str]:
    if docstring_src_cls is None:
        return None

    src = getattr(docstring_src_cls, member_name, None)
    if src is None:
        return None

    doc = getattr(src, "__doc__", None)
    if isinstance(src, property):
        # some things like SeriesGroupBy.unique are generated.
        src = src.fget
        if not doc:
            doc = getattr(src, "__doc__", None)

    # pandas DataFrame/Series sometimes override methods without setting __doc__.
    if doc is None and docstring_src_cls.__name__ in {"DataFrame", "Series"}:
        for cls in docstring_src_cls.mro():
            src = getattr(cls, member_name, None)
            if src is not None and hasattr(src, "__doc__"):
                doc = src.__doc__
    return doc


def attach_module_callable_docstring(
    c: Callable,
    docstring_src_module: ModuleType,
    docstring_src: Optional[Callable],
    fallback_warning: bool = False,
) -> Callable:
    """
    Attach docstring to functions and constructors.
    """

    if docstring_src is None:
        c.__doc__ = ""
        return c

    doc = getattr(docstring_src, "__doc__", None)
    doc = skip_doctest(doc)
    doc = add_version_disclaimer(doc, docstring_src_module)
    doc = add_arg_disclaimer(docstring_src, c, doc)
    doc = add_docstring_disclaimer(docstring_src_module, None, doc, fallback_warning)
    c.__doc__ = "" if doc is None else doc
    return c


def attach_cls_member_docstring(
    member: Any,
    member_name: str,
    data_type: Optional[DataType] = None,
    docstring_src_module: Optional[ModuleType] = None,
    docstring_src_cls: Optional[Type] = None,
    fallback_warning: bool = False,
) -> Any:
    """
    Attach docstring to class members.
    """

    if (
        docstring_src_module is None
        and docstring_src_cls is None
        and data_type is not None
    ):
        docstring_src_module, docstring_src_cls = _DATA_TYPE_TO_DOCSTRING_SRC.get(
            data_type, (None, None)
        )
    doc = gen_member_docstring(docstring_src_cls, member_name)
    doc = skip_doctest(doc)
    doc = add_version_disclaimer(doc, docstring_src_module)
    docstring_src = getattr(docstring_src_cls, member_name, None)
    doc = add_arg_disclaimer(docstring_src, member, doc)
    doc = add_docstring_disclaimer(
        docstring_src_module, docstring_src_cls, doc, fallback_warning
    )
    member.__doc__ = "" if doc is None else doc
    return member



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas

from ..core.adapter import (
    MarsDataFrameAt,
    MarsDataFrameIat,
    MarsDataFrameIloc,
    MarsDataFrameLoc,
    MarsGetItemProxy,
    register_converter,
)
from ..core.utils.docstring import attach_module_callable_docstring


@register_converter(from_cls_list=[MarsDataFrameAt])
class DataFrameAt(MarsGetItemProxy):
    pass


attach_module_callable_docstring(DataFrameAt, pandas, pandas.core.indexing._AtIndexer)


@register_converter(from_cls_list=[MarsDataFrameIat])
class DataFrameIat(MarsGetItemProxy):
    pass


attach_module_callable_docstring(DataFrameIat, pandas, pandas.core.indexing._iAtIndexer)


@register_converter(from_cls_list=[MarsDataFrameLoc])
class DataFrameLoc(MarsGetItemProxy):
    pass


attach_module_callable_docstring(DataFrameLoc, pandas, pandas.core.indexing._LocIndexer)


@register_converter(from_cls_list=[MarsDataFrameIloc])
class DataFrameIloc(MarsGetItemProxy):
    pass


attach_module_callable_docstring(
    DataFrameIloc, pandas, pandas.core.indexing._iLocIndexer
)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Any, Callable, Dict, Set

from .....core import TileableData
from .....dataframe import NamedAgg
from .....dataframe.arithmetic.core import DataFrameBinOp, DataFrameUnaryOp
from .....dataframe.core import BaseDataFrameData, BaseSeriesData
from .....dataframe.groupby.aggregation import DataFrameGroupByAgg
from .....dataframe.indexing.getitem import DataFrameIndex
from .....dataframe.indexing.setitem import DataFrameSetitem
from .....dataframe.merge import DataFrameMerge
from .....typing import OperandType
from .utils import get_cols_exclude_index


class InputColumnSelector:
    _OP_TO_SELECT_FUNCTION = {}

    @staticmethod
    def select_all_input_columns(
        tileable_data: TileableData, _required_cols: Set[Any]
    ) -> Dict[TileableData, Set[Any]]:
        ret = {}
        for inp in tileable_data.op.inputs:
            if isinstance(inp, BaseDataFrameData):
                ret[inp] = set(inp.dtypes.index)
            elif isinstance(inp, BaseSeriesData):
                ret[inp] = {inp.name}
        return ret

    @staticmethod
    def select_required_input_columns(
        tileable_data: TileableData, required_cols: Set[Any]
    ) -> Dict[TileableData, Set[Any]]:
        ret = {}
        for inp in tileable_data.op.inputs:
            if isinstance(inp, BaseDataFrameData):
                ret[inp] = required_cols.intersection(set(inp.dtypes.index))
            elif isinstance(inp, BaseSeriesData):
                ret[inp] = {inp.name}
        return ret

    @classmethod
    def register(
        cls,
        op_cls: OperandType,
        func: Callable[[TileableData, Set[Any]], Dict[TileableData, Set[Any]]],
        replace: bool = False,
    ) -> None:
        if op_cls not in cls._OP_TO_SELECT_FUNCTION or replace:
            cls._OP_TO_SELECT_FUNCTION[op_cls] = func
        else:
            raise ValueError(f"key {op_cls} exists.")

    @classmethod
    def unregister(cls, op_cls: OperandType) -> None:
        if op_cls in cls._OP_TO_SELECT_FUNCTION:
            del cls._OP_TO_SELECT_FUNCTION[op_cls]

    @classmethod
    def select(
        cls, tileable_data: TileableData, required_cols: Set[Any]
    ) -> Dict[TileableData, Set[Any]]:
        """
        Get the column pruning results of given tileable data.

        Parameters
        ----------
        tileable_data : TileableData
            The tileable data to be processed.
        required_cols: List[Any]
            Names of columns required by the successors of the given tileable data. If required_cols is None, all the
            input columns will be selected.
        Returns
        -------
        Dict[TileableData: List[Any]]
            A dictionary that represents the column pruning results. For every key-value pairs in the dictionary, the
            key is a predecessor of the given tileable data, and the value is a list of column names that the given
            tileable data depends on.
        """
        if required_cols is None:
            return cls.select_all_input_columns(tileable_data, set())

        op_type = type(tileable_data.op)
        if op_type in cls._OP_TO_SELECT_FUNCTION:
            return cls._OP_TO_SELECT_FUNCTION[op_type](tileable_data, required_cols)
        for op_cls in op_type.__mro__:
            if op_cls in cls._OP_TO_SELECT_FUNCTION:
                cls._OP_TO_SELECT_FUNCTION[op_type] = cls._OP_TO_SELECT_FUNCTION[op_cls]
                return cls._OP_TO_SELECT_FUNCTION[op_cls](tileable_data, required_cols)
        return cls.select_all_input_columns(tileable_data, required_cols)


def register_selector(op_type: OperandType) -> Callable:
    def wrap(selector_func: Callable):
        InputColumnSelector.register(op_type, selector_func)
        return selector_func

    return wrap


@register_selector(DataFrameMerge)
def df_merge_select_function(
    tileable_data: TileableData, required_cols: Set[Any]
) -> Dict[TileableData, Set[Any]]:
    op: DataFrameMerge = tileable_data.op
    assert len(op.inputs) == 2
    assert isinstance(op.inputs[0], BaseDataFrameData)
    assert isinstance(op.inputs[1], BaseDataFrameData)
    left_data: BaseDataFrameData = op.inputs[0]
    right_data: BaseDataFrameData = op.inputs[1]

    ret = defaultdict(set)
    for df, suffix in zip([left_data, right_data], op.suffixes):
        for col in df.dtypes.index:
            if col in required_cols:
                ret[df].add(col)
            else:
                # TODO: this does not work when col is a tuple.
                suffix_col = str(col) + suffix
                if suffix_col in required_cols:
                    ret[df].add(col)
                    # The column in the other dataframe has to be selected as well. Otherwise, in
                    # the runtime, there will not be a column with suffix.
                    other_data = right_data if df is left_data else left_data
                    ret[other_data].add(col)

    if op.on is not None:
        ret[left_data].update(get_cols_exclude_index(left_data, op.on))
        ret[right_data].update(get_cols_exclude_index(right_data, op.on))
    if op.left_on is not None:
        ret[left_data].update(get_cols_exclude_index(left_data, op.left_on))
    if op.right_on is not None:
        ret[right_data].update(get_cols_exclude_index(right_data, op.right_on))

    return ret


@register_selector(DataFrameGroupByAgg)
def df_groupby_agg_select_function(
    tileable_data: TileableData, required_cols: Set[Any]
) -> Dict[TileableData, Set[Any]]:
    op: DataFrameGroupByAgg = tileable_data.op
    assert isinstance(op.inputs[0], (BaseDataFrameData, BaseSeriesData))
    inp: BaseDataFrameData = op.inputs[0]
    by = op.groupby_params["by"]
    selection = op.groupby_params.get("selection", None)
    raw_func = op.raw_func

    ret = {}
    # group by a series
    groupby_series = False
    if isinstance(by, list) and all([isinstance(_by, BaseSeriesData) for _by in by]):
        groupby_series = True
        for _by in by:
            ret[_by] = {_by.name}

    if isinstance(inp, BaseSeriesData):
        ret[inp] = {inp.name}
    else:
        selected_cols = set()
        # group by keys should be included
        if not groupby_series:
            selected_cols.update(get_cols_exclude_index(inp, by))
        # add agg columns
        if op.raw_func is not None:
            if op.raw_func == "size":
                # special for size, its return value is always series
                pass
            elif isinstance(raw_func, dict):
                selected_cols.update(set(raw_func.keys()))
            else:
                # no specified agg columns
                # required_cols should always be a subset of selection
                for col in required_cols:
                    # col is a tuple when required col is a MultiIndex
                    if isinstance(col, tuple):
                        for c in col:
                            selected_cols.add(c)
                    selected_cols.add(col)
                if selection is not None:
                    if isinstance(selection, (list, tuple)):
                        selected_cols.update(set(selection))
                    else:
                        selected_cols.add(selection)
        elif op.raw_func_kw:
            # add renamed columns
            for _, origin in op.raw_func_kw.items():
                if isinstance(origin, NamedAgg):
                    selected_cols.add(origin.column)
                else:
                    assert isinstance(origin, tuple)
                    selected_cols.add(origin[0])

        ret[inp] = selected_cols.intersection(inp.dtypes.index)
    return ret


@register_selector(DataFrameSetitem)
def df_setitem_select_function(
    tileable_data: TileableData, required_cols: Set[Any]
) -> Dict[TileableData, Set[Any]]:
    if len(tileable_data.inputs) == 1:
        # if value is not a Mars object, return required input columns
        return InputColumnSelector.select_required_input_columns(
            tileable_data, required_cols
        )
    else:
        df, value = tileable_data.inputs
        ret = {df: required_cols.intersection(set(df.dtypes.index))}
        # if value is a Mars object, return all its columns so that setitem can be executed
        if isinstance(value, BaseDataFrameData):
            value_cols = set(value.dtypes.index)
            ret[value] = value_cols
        elif isinstance(value, BaseSeriesData):
            value_cols = {value.name}
            ret[value] = value_cols
        return ret


@register_selector(DataFrameIndex)
def df_getitem_select_function(
    tileable_data: TileableData, required_cols: Set[Any]
) -> Dict[TileableData, Set[Any]]:
    if tileable_data.op.col_names:
        return InputColumnSelector.select_required_input_columns(
            tileable_data, required_cols
        )
    else:
        return InputColumnSelector.select_all_input_columns(
            tileable_data, required_cols
        )


SELECT_REQUIRED_OP_TYPES = [DataFrameBinOp, DataFrameUnaryOp]
for op_type in SELECT_REQUIRED_OP_TYPES:
    InputColumnSelector.register(
        op_type, InputColumnSelector.select_required_input_columns
    )



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from . import flatiter
from .core import (
    MARS_TENSOR_CALLABLES,
    MARS_TENSOR_FFT_CALLABLES,
    MARS_TENSOR_LINALG_CALLABLES,
    MARS_TENSOR_MAGIC_METHODS,
    MARS_TENSOR_OBJECTS,
    MARS_TENSOR_RANDOM_CALLABLES,
    MARS_TENSOR_SPECIAL_CALLABLES,
)


def _install():
    from ...core.adapter import (
        MARS_TENSOR_TYPE,
        collect_cls_members,
        register_data_members,
        wrap_magic_method,
    )
    from ...core.data import DataRef, DataType

    for method in MARS_TENSOR_MAGIC_METHODS:
        setattr(DataRef, method, wrap_magic_method(method))

    for cls in MARS_TENSOR_TYPE:
        register_data_members(
            DataType.tensor, collect_cls_members(cls, DataType.tensor)
        )



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Type

from ....core import ChunkGraph
from ....typing import OperandType
from ..core import OperandBasedOptimizationRule, OptimizationRecords, Optimizer


class ChunkOptimizer(Optimizer):
    """
    Chunk Optimizer
    """


def register_operand_based_optimization_rule(op_types: List[Type[OperandType]]):
    def wrap(rule_type: Type[OperandBasedOptimizationRule]):
        for op_type in op_types:
            rule_type.register_operand(op_type)
        ChunkOptimizer.register_rule(rule_type)

    return wrap


def optimize(chunk_graph: ChunkGraph) -> OptimizationRecords:
    return ChunkOptimizer.optimize(chunk_graph)



# Copyright 2022-2023 XProbe Inc.
# derived from copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import bisect
import os
from typing import Any, List, Optional

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core.operand import MapReduceOperand, OperandStage
from ...serialization.serializables import (
    AnyField,
    BoolField,
    Int32Field,
    ListField,
    StringField,
)
from ...tensor.base.psrs import PSRSOperandMixin
from ...utils import calc_nsplits, lazy_import
from ..core import IndexValue, OutputType
from ..operands import DataFrameOperand, DataFrameOperandMixin, DataFrameShuffleProxy
from ..utils import is_cudf, parse_index, standardize_range_index

cudf = lazy_import("cudf")

_PSRS_DISTINCT_COL = "__PSRS_TMP_DISTINCT_COL"


class _Largest:
    """
    This util class resolve TypeError when
    comparing strings with None values
    """

    def __lt__(self, other):
        return False

    def __gt__(self, other):
        return self is not other


_largest = _Largest()


class _ReversedValue:
    def __init__(self, value):
        self._value = value

    def __lt__(self, other):
        if type(other) is _ReversedValue:
            # may happen when call searchsorted
            return self._value >= other._value
        return self._value >= other

    def __gt__(self, other):
        return self._value <= other

    def __repr__(self):
        return repr(self._value)


class DataFramePSRSOperandMixin(DataFrameOperandMixin, PSRSOperandMixin):
    @classmethod
    def _collect_op_properties(cls, op):
        from .sort_values import DataFrameSortValues

        if isinstance(op, DataFrameSortValues):
            properties = dict(
                sort_type="sort_values",
                axis=op.axis,
                by=op.by,
                ascending=op.ascending,
                inplace=op.inplace,
                na_position=op.na_position,
                gpu=op.is_gpu(),
            )
        else:
            properties = dict(
                sort_type="sort_index",
                axis=op.axis,
                level=op.level,
                ascending=op.ascending,
                inplace=op.inplace,
                na_position=op.na_position,
                sort_remaining=op.sort_remaining,
                gpu=op.is_gpu(),
            )
        return properties

    @classmethod
    def local_sort_and_regular_sample(
        cls, op, in_data, axis_chunk_shape, axis_offsets, out_idx
    ):
        # stage 1: local sort and regular samples collected
        sorted_chunks, indices_chunks, sampled_chunks = [], [], []
        for i in range(axis_chunk_shape):
            in_chunk = in_data.chunks[i]
            kind = None if op.psrs_kinds is None else op.psrs_kinds[0]
            chunk_op = DataFramePSRSSortRegularSample(
                kind=kind,
                n_partition=axis_chunk_shape,
                output_types=op.output_types,
                **cls._collect_op_properties(op),
            )
            kws = []
            sort_shape = in_chunk.shape
            kws.append(
                {
                    "shape": sort_shape,
                    "index_value": in_chunk.index_value,
                    "index": in_chunk.index,
                }
            )
            if chunk_op.sort_type == "sort_values":
                sampled_shape = (
                    (axis_chunk_shape, len(op.by)) if op.by else (axis_chunk_shape,)
                )
            else:
                sampled_shape = (
                    (axis_chunk_shape, sort_shape[1])
                    if len(sort_shape) == 2
                    else (axis_chunk_shape,)
                )
            kws.append(
                {
                    "shape": sampled_shape,
                    "index_value": in_chunk.index_value,
                    "index": (i,),
                    "type": "regular_sampled",
                }
            )
            if op.outputs[0].ndim == 2:
                kws[0].update(
                    {"columns_value": in_chunk.columns_value, "dtypes": in_chunk.dtypes}
                )
                kws[1].update(
                    {"columns_value": in_chunk.columns_value, "dtypes": in_chunk.dtypes}
                )
            else:
                kws[0].update(({"dtype": in_chunk.dtype, "name": in_chunk.name}))
                kws[1].update({"dtype": in_chunk.dtype})

            chunks = chunk_op.new_chunks([in_chunk], kws=kws, output_limit=len(kws))
            sort_chunk, sampled_chunk = chunks
            sorted_chunks.append(sort_chunk)
            sampled_chunks.append(sampled_chunk)
        return sorted_chunks, indices_chunks, sampled_chunks

    @classmethod
    def concat_and_pivot(
        cls, op, axis_chunk_shape, out_idx, sorted_chunks, sampled_chunks
    ):
        from .sort_values import DataFrameSortValues

        # stage 2: gather and merge samples, choose and broadcast p-1 pivots
        kind = None if op.psrs_kinds is None else op.psrs_kinds[1]
        if isinstance(op, DataFrameSortValues):
            output_types = op.output_types
        else:
            output_types = [OutputType.index]
        concat_pivot_op = DataFramePSRSConcatPivot(
            kind=kind,
            n_partition=axis_chunk_shape,
            output_types=output_types,
            **cls._collect_op_properties(op),
        )
        concat_pivot_shape = (
            sorted_chunks[0].shape[: op.axis]
            + (axis_chunk_shape - 1,)
            + sorted_chunks[0].shape[op.axis + 1 :]
        )
        concat_pivot_index = out_idx[: op.axis] + (0,) + out_idx[op.axis :]
        concat_pivot_chunk = concat_pivot_op.new_chunk(
            sampled_chunks,
            shape=concat_pivot_shape,
            index=concat_pivot_index,
        )
        return concat_pivot_chunk

    @classmethod
    def partition_local_data(
        cls, op, axis_chunk_shape, sorted_chunks, indices_chunks, concat_pivot_chunk
    ):
        # stage 3: Local data is partitioned
        partition_chunks = []
        length = len(sorted_chunks)
        for i in range(length):
            chunk_inputs = [sorted_chunks[i], concat_pivot_chunk]
            partition_shuffle_map = DataFramePSRSShuffle(
                n_partition=axis_chunk_shape,
                stage=OperandStage.map,
                output_types=op.output_types,
                **cls._collect_op_properties(op),
            )
            if isinstance(chunk_inputs[0].index_value.value, IndexValue.RangeIndex):
                index_value = parse_index(pd.Index([], dtype=np.int64))
            else:
                index_value = chunk_inputs[0].index_value
            kw = dict(
                shape=chunk_inputs[0].shape,
                index=chunk_inputs[0].index,
                index_value=index_value,
            )
            if op.outputs[0].ndim == 2:
                kw.update(
                    dict(
                        columns_value=chunk_inputs[0].columns_value,
                        dtypes=chunk_inputs[0].dtypes,
                    )
                )
            else:
                kw.update(dict(dtype=chunk_inputs[0].dtype, name=chunk_inputs[0].name))
            partition_chunk = partition_shuffle_map.new_chunk(chunk_inputs, **kw)
            partition_chunks.append(partition_chunk)
        return partition_chunks

    @classmethod
    def partition_merge_data(
        cls, op, need_align, return_value, partition_chunks, proxy_chunk
    ):
        # stage 4: all *ith* classes are gathered and merged
        partition_sort_chunks, partition_indices_chunks, sort_info_chunks = [], [], []
        for i, partition_chunk in enumerate(partition_chunks):
            kind = None if op.psrs_kinds is None else op.psrs_kinds[2]
            partition_shuffle_reduce = DataFramePSRSShuffle(
                stage=OperandStage.reduce,
                kind=kind,
                reducer_index=(i,),
                n_reducers=len(partition_chunks),
                output_types=op.output_types,
                **cls._collect_op_properties(op),
            )
            chunk_shape = list(partition_chunk.shape)
            chunk_shape[op.axis] = np.nan

            kw = dict(
                shape=tuple(chunk_shape),
                index=partition_chunk.index,
                index_value=partition_chunk.index_value,
            )
            if op.outputs[0].ndim == 2:
                kw.update(
                    dict(
                        columns_value=partition_chunk.columns_value,
                        dtypes=partition_chunk.dtypes,
                    )
                )
            else:
                kw.update(dict(dtype=partition_chunk.dtype, name=partition_chunk.name))
            cs = partition_shuffle_reduce.new_chunks([proxy_chunk], **kw)

            partition_sort_chunks.append(cs[0])
        return partition_sort_chunks, partition_indices_chunks, sort_info_chunks

    @classmethod
    def _tile_psrs(cls, op, in_data):
        out = op.outputs[0]
        in_df, axis_chunk_shape, _, _ = yield from cls.preprocess(op, in_data=in_data)

        # stage 1: local sort and regular samples collected
        sorted_chunks, _, sampled_chunks = cls.local_sort_and_regular_sample(
            op, in_df, axis_chunk_shape, None, None
        )

        # stage 2: gather and merge samples, choose and broadcast p-1 pivots
        concat_pivot_chunk = cls.concat_and_pivot(
            op,
            axis_chunk_shape,
            (0,) if in_df.ndim == 2 else (),
            sorted_chunks,
            sampled_chunks,
        )

        # stage 3: Local data is partitioned
        partition_chunks = cls.partition_local_data(
            op, axis_chunk_shape, sorted_chunks, None, concat_pivot_chunk
        )

        proxy_chunk = DataFrameShuffleProxy(output_types=op.output_types).new_chunk(
            partition_chunks, shape=()
        )

        # stage 4: all *ith* classes are gathered and merged
        partition_sort_chunks = cls.partition_merge_data(
            op, False, None, partition_chunks, proxy_chunk
        )[0]

        if op.ignore_index:
            yield partition_sort_chunks
            chunks = standardize_range_index(partition_sort_chunks, axis=op.axis)
        else:
            chunks = partition_sort_chunks

        nsplits = calc_nsplits({c.index: c.shape for c in chunks})
        if op.outputs[0].ndim == 2:
            new_op = op.copy()
            return new_op.new_dataframes(
                op.inputs,
                shape=out.shape,
                chunks=chunks,
                nsplits=nsplits,
                index_value=out.index_value,
                columns_value=out.columns_value,
                dtypes=out.dtypes,
            )
        else:
            new_op = op.copy()
            return new_op.new_seriess(
                op.inputs,
                shape=out.shape,
                chunks=chunks,
                nsplits=nsplits,
                index_value=out.index_value,
                dtype=out.dtype,
                name=out.name,
            )


def execute_sort_values(data, op, inplace=None, by=None):
    if inplace is None:
        inplace = op.inplace
    # ignore_index is new in Pandas version 1.0.0.
    ignore_index = getattr(op, "ignore_index", False)
    if isinstance(data, (pd.DataFrame, pd.Series)):
        kwargs = dict(
            axis=op.axis,
            ascending=op.ascending,
            ignore_index=ignore_index,
            na_position=op.na_position,
            kind=op.kind,
        )
        if isinstance(data, pd.DataFrame):
            kwargs["by"] = by if by is not None else op.by
        if inplace:
            kwargs["inplace"] = True
            try:
                data.sort_values(**kwargs)
            except TypeError:  # pragma: no cover
                kwargs.pop("ignore_index", None)
                data.sort_values(**kwargs)
            return data
        else:
            try:
                return data.sort_values(**kwargs)
            except TypeError:  # pragma: no cover
                kwargs.pop("ignore_index", None)
                return data.sort_values(**kwargs)

    else:  # pragma: no cover
        # cudf doesn't support axis and kind
        if isinstance(data, cudf.DataFrame):
            return data.sort_values(
                op.by, ascending=op.ascending, na_position=op.na_position
            )
        else:
            return data.sort_values(ascending=op.ascending, na_position=op.na_position)


def execute_sort_index(data, op, inplace=None):
    if inplace is None:
        inplace = op.inplace
    # ignore_index is new in Pandas version 1.0.0.
    ignore_index = getattr(op, "ignore_index", False)
    if isinstance(data, (pd.DataFrame, pd.Series)):
        kwargs = dict(
            level=op.level,
            ascending=op.ascending,
            ignore_index=ignore_index,
            na_position=op.na_position,
            kind=op.kind,
            sort_remaining=op.sort_remaining,
        )
        if inplace:
            kwargs["inplace"] = True
            try:
                data.sort_index(**kwargs)
            except TypeError:  # pragma: no cover
                kwargs.pop("ignore_index", None)
                data.sort_index(**kwargs)
            return data
        else:
            try:
                return data.sort_index(**kwargs)
            except TypeError:  # pragma: no cover
                kwargs.pop("ignore_index", None)
                return data.sort_index(**kwargs)

    else:  # pragma: no cover
        # cudf only support ascending
        return data.sort_index(ascending=op.ascending)


class DataFramePSRSChunkOperand(DataFrameOperand):
    # sort type could be 'sort_values' or 'sort_index'
    sort_type = StringField("sort_type")

    axis = Int32Field("axis")
    by = ListField("by", default=None)
    ascending = AnyField("ascending")
    inplace = BoolField("inplace")
    kind = StringField("kind")
    na_position = StringField("na_position")

    # for sort_index
    level = ListField("level")
    sort_remaining = BoolField("sort_remaining")

    n_partition = Int32Field("n_partition")

    def __init__(self, output_types=None, **kw):
        super().__init__(_output_types=output_types, **kw)


class DataFramePSRSSortRegularSample(DataFramePSRSChunkOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.PSRS_SORT_REGULAR_SMAPLE

    @property
    def output_limit(self):
        return 2

    @classmethod
    def execute(cls, ctx, op):
        a = ctx[op.inputs[0].key]
        xdf = pd if isinstance(a, (pd.DataFrame, pd.Series)) else cudf

        if len(a) == 0:
            # when chunk is empty, return the empty chunk itself
            ctx[op.outputs[0].key] = ctx[op.outputs[-1].key] = a
            return

        if op.sort_type == "sort_values":
            ctx[op.outputs[0].key] = res = execute_sort_values(a, op)
        else:
            ctx[op.outputs[0].key] = res = execute_sort_index(a, op)

        by = op.by
        add_distinct_col = bool(int(os.environ.get("PSRS_DISTINCT_COL", "0")))
        if (
            add_distinct_col
            and isinstance(a, xdf.DataFrame)
            and op.sort_type == "sort_values"
        ):
            # when running under distributed mode, we introduce an extra column
            # to make sure pivots are distinct
            chunk_idx = op.inputs[0].index[0]
            distinct_col = (
                _PSRS_DISTINCT_COL
                if a.columns.nlevels == 1
                else (_PSRS_DISTINCT_COL,) + ("",) * (a.columns.nlevels - 1)
            )
            res[distinct_col] = np.arange(
                chunk_idx << 32, (chunk_idx << 32) + len(a), dtype=np.int64
            )
            by = list(by) + [distinct_col]

        n = op.n_partition
        if op.sort_type == "sort_values" and a.shape[op.axis] < n:
            num = n // a.shape[op.axis] + 1
            res = execute_sort_values(xdf.concat([res] * num), op, by=by)

        w = res.shape[op.axis] * 1.0 / (n + 1)
        slc = np.linspace(
            max(w - 1, 0), res.shape[op.axis] - 1, num=n, endpoint=False
        ).astype(int)
        if op.axis == 1:
            slc = (slice(None), slc)
        if op.sort_type == "sort_values":
            # do regular sample
            if op.by is not None:
                ctx[op.outputs[-1].key] = res[by].iloc[slc]
            else:
                ctx[op.outputs[-1].key] = res.iloc[slc]
        else:
            # do regular sample
            ctx[op.outputs[-1].key] = res.iloc[slc]


class DataFramePSRSConcatPivot(DataFramePSRSChunkOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.PSRS_CONCAT_PIVOT

    @property
    def output_limit(self):
        return 1

    @classmethod
    def execute(cls, ctx, op):
        inputs = [ctx[c.key] for c in op.inputs if len(ctx[c.key]) > 0]
        if len(inputs) == 0:
            # corner case: nothing sampled, we need to do nothing
            ctx[op.outputs[-1].key] = ctx[op.inputs[0].key]
            return

        xdf = pd if isinstance(inputs[0], (pd.DataFrame, pd.Series)) else cudf

        a = xdf.concat(inputs, axis=op.axis)
        p = len(inputs)
        assert a.shape[op.axis] == p * len(op.inputs)

        slc = np.linspace(
            p - 1, a.shape[op.axis] - 1, num=len(op.inputs) - 1, endpoint=False
        ).astype(int)
        if op.axis == 1:
            slc = (slice(None), slc)
        if op.sort_type == "sort_values":
            a = execute_sort_values(a, op, inplace=False)
            ctx[op.outputs[-1].key] = a.iloc[slc]
        else:
            a = execute_sort_index(a, op, inplace=False)
            ctx[op.outputs[-1].key] = a.index[slc]


class DataFramePSRSShuffle(MapReduceOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.PSRS_SHUFFLE

    sort_type = StringField("sort_type")

    # for shuffle map
    axis = Int32Field("axis")
    by = ListField("by")
    ascending = AnyField("ascending")
    inplace = BoolField("inplace")
    na_position = StringField("na_position")
    n_partition = Int32Field("n_partition")

    # for sort_index
    level = ListField("level")
    sort_remaining = BoolField("sort_remaining")

    # for shuffle reduce
    kind = StringField("kind")

    def __init__(self, output_types=None, **kw):
        super().__init__(_output_types=output_types, **kw)

    @property
    def output_limit(self):
        return 1

    @staticmethod
    def _calc_poses(src_cols, pivots, ascending=True):
        # The pivots are immutable if it is got from shared memory, e.g. Ray object store.
        # Pandas < 1.4 has item setting bug and pandas >= 1.4 has fixed it.
        #
        # Here, almost all the cases that the pivots are got from shared memory.
        #
        # `pivots[col] = -pivots[col]` will automatically replace the col with a new copy
        # `-pivots[col]` in pandas >= 1.4, but it will try to inplace set col in pandas < 1.4
        #
        # So, we use assign here to walk around incorrect inplace set item bug in pandas < 1.4.
        # Please refer to: https://github.com/mars-project/mars/issues/3215
        # related issue: https://github.com/pandas-dev/pandas/pull/43406
        copy_cols = {}
        if isinstance(ascending, list):
            for asc, col in zip(ascending, pivots.columns):
                # Make pivots available to use ascending order when mixed order specified
                if not asc:
                    if pd.api.types.is_numeric_dtype(pivots.dtypes[col]):
                        # for numeric dtypes, convert to negative is more efficient
                        copy_cols[col] = -pivots[col]
                        src_cols[col] = -src_cols[col]
                    else:
                        # for other types, convert to ReversedValue
                        copy_cols[col] = pivots[col].map(
                            lambda x: x
                            if type(x) is _ReversedValue
                            else _ReversedValue(x)
                        )
            ascending = True

        if copy_cols:
            pivots = pivots.assign(**copy_cols)

        records = src_cols.to_records(index=False)
        p_records = pivots.to_records(index=False)
        if ascending:
            poses = records.searchsorted(p_records, side="right")
        else:
            poses = len(records) - records[::-1].searchsorted(p_records, side="right")
        del records, p_records
        return poses

    @classmethod
    def _execute_dataframe_map(cls, ctx, op):
        a, pivots = [ctx[c.key] for c in op.inputs]
        out = op.outputs[0]

        if len(a) == 0:
            # when the chunk is empty, no slices can be produced
            for i in range(op.n_partition):
                ctx[out.key, (i,)] = a
            return

        # use numpy.searchsorted to find split positions.
        by = op.by

        distinct_col = (
            _PSRS_DISTINCT_COL
            if a.columns.nlevels == 1
            else (_PSRS_DISTINCT_COL,) + ("",) * (a.columns.nlevels - 1)
        )
        if distinct_col in a.columns:
            by = list(by) + [distinct_col]

        try:
            poses = cls._calc_poses(a[by], pivots, op.ascending)
        except TypeError:
            poses = cls._calc_poses(
                a[by].fillna(_largest), pivots.fillna(_largest), op.ascending
            )

        poses = (None,) + tuple(poses) + (None,)
        for i in range(op.n_partition):
            values = a.iloc[poses[i] : poses[i + 1]]
            if is_cudf(values):  # pragma: no cover
                values = values.copy()
            ctx[out.key, (i,)] = values

    @classmethod
    def _calc_series_poses(cls, s, pivots, ascending=True):
        if ascending:
            poses = s.searchsorted(pivots, side="right")
        else:
            poses = len(s) - s.iloc[::-1].searchsorted(pivots, side="right")
        return poses

    @classmethod
    def _execute_series_map(cls, ctx, op):
        a, pivots = [ctx[c.key] for c in op.inputs]
        out = op.outputs[0]

        if len(a) == 0:
            # when the chunk is empty, no slices can be produced
            for i in range(op.n_partition):
                ctx[out.key, (i,)] = a
            return

        if isinstance(a, pd.Series):
            try:
                poses = cls._calc_series_poses(a, pivots, ascending=op.ascending)
            except TypeError:
                filled_a = a.fillna(_largest)
                filled_pivots = pivots.fillna(_largest)
                poses = cls._calc_series_poses(
                    filled_a, filled_pivots, ascending=op.ascending
                )
            poses = (None,) + tuple(poses) + (None,)
            for i in range(op.n_partition):
                values = a.iloc[poses[i] : poses[i + 1]]
                ctx[out.key, (i,)] = values

    @classmethod
    def _calc_pivots_poses(
        cls,
        index: Any,
        pivots: Any,
        level: Optional[List] = None,
        sort_remaining: bool = True,
    ):
        if isinstance(index, pd.MultiIndex) and isinstance(pivots, pd.MultiIndex):
            is_multi_index = True
        elif not isinstance(index, pd.MultiIndex) and not isinstance(
            pivots, pd.MultiIndex
        ):
            is_multi_index = False
        else:  # pragma: no cover
            raise TypeError(
                f"Invalid types for index and pivots: index is {type(index)}, while pivots is {type(pivots)}."
            )

        if not is_multi_index:
            return index.searchsorted(list(pivots), side="right")

        if index.nlevels != pivots.nlevels:  # pragma: no cover
            raise ValueError(
                f"Inconsistent levels: index has {index.nlevels} levels, while pivots has {pivots.nlevels} levels."
            )

        if isinstance(level, list) and len(level) == 1 and not sort_remaining:
            return index.get_level_values(level[0]).searchsorted(
                list(pivots.get_level_values(level[0])), side="right"
            )
        else:
            if level is None:
                index_list = list(index)
                pivots_list = list(pivots)
            elif isinstance(level, list):
                index_from_level = []
                pivots_from_level = []
                for l in level:
                    index_from_level.append(index.get_level_values(l))
                    pivots_from_level.append(pivots.get_level_values(l))
                if sort_remaining:
                    if len(level) < index.nlevels:
                        remain_indexes = index.droplevel(level)
                        remain_pivots = pivots.droplevel(level)
                        for l in range(remain_indexes.nlevels):
                            index_from_level.append(remain_indexes.get_level_values(l))
                        for l in range(remain_pivots.nlevels):
                            pivots_from_level.append(remain_pivots.get_level_values(l))
                index_list = list(zip(*index_from_level))
                pivots_list = list(zip(*pivots_from_level))
            else:  # pragma: no cover
                raise TypeError(f"Invalid level type: {type(level)}")

            poses = []
            for pivot in pivots_list:
                poses.append(bisect.bisect_right(index_list, pivot))

            return np.array(poses)

    @classmethod
    def _execute_sort_index_map(cls, ctx, op):
        a, pivots = [ctx[c.key] for c in op.inputs]
        out = op.outputs[0]

        if op.ascending:
            poses = cls._calc_pivots_poses(a.index, pivots, op.level, op.sort_remaining)
        else:
            poses = len(a) - cls._calc_pivots_poses(
                a.index[::-1], pivots, op.level, op.sort_remaining
            )
        poses = (None,) + tuple(poses) + (None,)
        for i in range(op.n_partition):
            values = a.iloc[poses[i] : poses[i + 1]]
            ctx[out.key, (i,)] = values

    @classmethod
    def _execute_map(cls, ctx, op):
        a = [ctx[c.key] for c in op.inputs][0]
        if op.sort_type == "sort_values":
            if len(a.shape) == 2:
                # DataFrame type
                cls._execute_dataframe_map(ctx, op)
            else:
                # Series type
                cls._execute_series_map(ctx, op)
        else:
            cls._execute_sort_index_map(ctx, op)

    @classmethod
    def _execute_reduce(cls, ctx, op: "DataFramePSRSShuffle"):
        out_chunk = op.outputs[0]
        raw_inputs = list(op.iter_mapper_data(ctx, pop=False))

        xdf = pd if isinstance(raw_inputs[0], (pd.DataFrame, pd.Series)) else cudf
        if xdf is pd:
            concat_values = xdf.concat(raw_inputs, axis=op.axis, copy=False)
        else:
            concat_values = xdf.concat(raw_inputs, axis=op.axis)
        del raw_inputs[:]

        if isinstance(concat_values, xdf.DataFrame):
            concat_values.drop(
                _PSRS_DISTINCT_COL, axis=1, inplace=True, errors="ignore"
            )

            col_index_dtype = out_chunk.columns_value.to_pandas().dtype
            if concat_values.columns.dtype != col_index_dtype:
                concat_values.columns = concat_values.columns.astype(col_index_dtype)

        if op.sort_type == "sort_values":
            ctx[op.outputs[0].key] = execute_sort_values(concat_values, op)
        else:
            ctx[op.outputs[0].key] = execute_sort_index(concat_values, op)

    @classmethod
    def estimate_size(cls, ctx, op):
        super().estimate_size(ctx, op)
        result = ctx[op.outputs[0].key]
        if op.stage == OperandStage.reduce:
            ctx[op.outputs[0].key] = (result[0], result[1] * 1.5)
        else:
            ctx[op.outputs[0].key] = result

    @classmethod
    def execute(cls, ctx, op):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        else:
            cls._execute_reduce(ctx, op)



# Copyright 2022-2023 XProbe Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import xoscar as mo

from ...core import AbstractService
from ..task_info_collector import TaskInfoCollectorActor


class TaskWorkerService(AbstractService):
    async def start(self):
        profiling_config = self._config.get("profiling", dict())
        await mo.create_actor(
            TaskInfoCollectorActor,
            profiling_config,
            uid=TaskInfoCollectorActor.default_uid(),
            address=self._address,
        )

    async def stop(self):
        await mo.destroy_actor(
            mo.create_actor_ref(
                uid=TaskInfoCollectorActor.default_uid(), address=self._address
            )
        )
